# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import debugger as Debugger
from chromewhip.protocol import dom as DOM
from chromewhip.protocol import network as Network
from chromewhip.protocol import runtime as Runtime
from chromewhip.protocol import emulation as Emulation

# ResourceType: Resource type as it was perceived by the rendering engine.
ResourceType = str

# FrameId: Unique frame identifier.
FrameId = str

# Frame: Information about the Frame on the page.
class Frame(ChromeTypeBase):
    def __init__(self,
                 id: Union['str'],
                 loaderId: Union['Network.LoaderId'],
                 url: Union['str'],
                 securityOrigin: Union['str'],
                 mimeType: Union['str'],
                 parentId: Optional['str'] = None,
                 name: Optional['str'] = None,
                 unreachableUrl: Optional['str'] = None,
                 ):

        self.id = id
        self.parentId = parentId
        self.loaderId = loaderId
        self.name = name
        self.url = url
        self.securityOrigin = securityOrigin
        self.mimeType = mimeType
        self.unreachableUrl = unreachableUrl


# FrameResource: Information about the Resource on the page.
class FrameResource(ChromeTypeBase):
    def __init__(self,
                 url: Union['str'],
                 type: Union['ResourceType'],
                 mimeType: Union['str'],
                 lastModified: Optional['Network.TimeSinceEpoch'] = None,
                 contentSize: Optional['float'] = None,
                 failed: Optional['bool'] = None,
                 canceled: Optional['bool'] = None,
                 ):

        self.url = url
        self.type = type
        self.mimeType = mimeType
        self.lastModified = lastModified
        self.contentSize = contentSize
        self.failed = failed
        self.canceled = canceled


# FrameResourceTree: Information about the Frame hierarchy along with their cached resources.
class FrameResourceTree(ChromeTypeBase):
    def __init__(self,
                 frame: Union['Frame'],
                 resources: Union['[FrameResource]'],
                 childFrames: Optional['[FrameResourceTree]'] = None,
                 ):

        self.frame = frame
        self.childFrames = childFrames
        self.resources = resources


# FrameTree: Information about the Frame hierarchy.
class FrameTree(ChromeTypeBase):
    def __init__(self,
                 frame: Union['Frame'],
                 childFrames: Optional['[FrameTree]'] = None,
                 ):

        self.frame = frame
        self.childFrames = childFrames


# ScriptIdentifier: Unique script identifier.
ScriptIdentifier = str

# TransitionType: Transition type.
TransitionType = str

# NavigationEntry: Navigation history entry.
class NavigationEntry(ChromeTypeBase):
    def __init__(self,
                 id: Union['int'],
                 url: Union['str'],
                 userTypedURL: Union['str'],
                 title: Union['str'],
                 transitionType: Union['TransitionType'],
                 ):

        self.id = id
        self.url = url
        self.userTypedURL = userTypedURL
        self.title = title
        self.transitionType = transitionType


# ScreencastFrameMetadata: Screencast frame metadata.
class ScreencastFrameMetadata(ChromeTypeBase):
    def __init__(self,
                 offsetTop: Union['float'],
                 pageScaleFactor: Union['float'],
                 deviceWidth: Union['float'],
                 deviceHeight: Union['float'],
                 scrollOffsetX: Union['float'],
                 scrollOffsetY: Union['float'],
                 timestamp: Optional['Network.TimeSinceEpoch'] = None,
                 ):

        self.offsetTop = offsetTop
        self.pageScaleFactor = pageScaleFactor
        self.deviceWidth = deviceWidth
        self.deviceHeight = deviceHeight
        self.scrollOffsetX = scrollOffsetX
        self.scrollOffsetY = scrollOffsetY
        self.timestamp = timestamp


# DialogType: Javascript dialog type.
DialogType = str

# AppManifestError: Error while paring app manifest.
class AppManifestError(ChromeTypeBase):
    def __init__(self,
                 message: Union['str'],
                 critical: Union['int'],
                 line: Union['int'],
                 column: Union['int'],
                 ):

        self.message = message
        self.critical = critical
        self.line = line
        self.column = column


# LayoutViewport: Layout viewport position and dimensions.
class LayoutViewport(ChromeTypeBase):
    def __init__(self,
                 pageX: Union['int'],
                 pageY: Union['int'],
                 clientWidth: Union['int'],
                 clientHeight: Union['int'],
                 ):

        self.pageX = pageX
        self.pageY = pageY
        self.clientWidth = clientWidth
        self.clientHeight = clientHeight


# VisualViewport: Visual viewport position, dimensions, and scale.
class VisualViewport(ChromeTypeBase):
    def __init__(self,
                 offsetX: Union['float'],
                 offsetY: Union['float'],
                 pageX: Union['float'],
                 pageY: Union['float'],
                 clientWidth: Union['float'],
                 clientHeight: Union['float'],
                 scale: Union['float'],
                 ):

        self.offsetX = offsetX
        self.offsetY = offsetY
        self.pageX = pageX
        self.pageY = pageY
        self.clientWidth = clientWidth
        self.clientHeight = clientHeight
        self.scale = scale


# Viewport: Viewport for capturing screenshot.
class Viewport(ChromeTypeBase):
    def __init__(self,
                 x: Union['float'],
                 y: Union['float'],
                 width: Union['float'],
                 height: Union['float'],
                 scale: Union['float'],
                 ):

        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.scale = scale


# FontFamilies: Generic font families collection.
class FontFamilies(ChromeTypeBase):
    def __init__(self,
                 standard: Optional['str'] = None,
                 fixed: Optional['str'] = None,
                 serif: Optional['str'] = None,
                 sansSerif: Optional['str'] = None,
                 cursive: Optional['str'] = None,
                 fantasy: Optional['str'] = None,
                 pictograph: Optional['str'] = None,
                 ):

        self.standard = standard
        self.fixed = fixed
        self.serif = serif
        self.sansSerif = sansSerif
        self.cursive = cursive
        self.fantasy = fantasy
        self.pictograph = pictograph


# FontSizes: Default font sizes.
class FontSizes(ChromeTypeBase):
    def __init__(self,
                 standard: Optional['int'] = None,
                 fixed: Optional['int'] = None,
                 ):

        self.standard = standard
        self.fixed = fixed


class Page(PayloadMixin):
    """ Actions and events related to the inspected page belong to the page domain.
    """
    @classmethod
    def addScriptToEvaluateOnLoad(cls,
                                  scriptSource: Union['str'],
                                  ):
        """Deprecated, please use addScriptToEvaluateOnNewDocument instead.
        :param scriptSource: 
        :type scriptSource: str
        """
        return (
            cls.build_send_payload("addScriptToEvaluateOnLoad", {
                "scriptSource": scriptSource,
            }),
            cls.convert_payload({
                "identifier": {
                    "class": ScriptIdentifier,
                    "optional": False
                },
            })
        )

    @classmethod
    def addScriptToEvaluateOnNewDocument(cls,
                                         source: Union['str'],
                                         ):
        """Evaluates given script in every frame upon creation (before loading frame's scripts).
        :param source: 
        :type source: str
        """
        return (
            cls.build_send_payload("addScriptToEvaluateOnNewDocument", {
                "source": source,
            }),
            cls.convert_payload({
                "identifier": {
                    "class": ScriptIdentifier,
                    "optional": False
                },
            })
        )

    @classmethod
    def bringToFront(cls):
        """Brings page to front (activates tab).
        """
        return (
            cls.build_send_payload("bringToFront", {
            }),
            None
        )

    @classmethod
    def captureScreenshot(cls,
                          format: Optional['str'] = None,
                          quality: Optional['int'] = None,
                          clip: Optional['Viewport'] = None,
                          fromSurface: Optional['bool'] = None,
                          ):
        """Capture page screenshot.
        :param format: Image compression format (defaults to png).
        :type format: str
        :param quality: Compression quality from range [0..100] (jpeg only).
        :type quality: int
        :param clip: Capture the screenshot of a given region only.
        :type clip: Viewport
        :param fromSurface: Capture the screenshot from the surface, rather than the view. Defaults to true.
        :type fromSurface: bool
        """
        return (
            cls.build_send_payload("captureScreenshot", {
                "format": format,
                "quality": quality,
                "clip": clip,
                "fromSurface": fromSurface,
            }),
            cls.convert_payload({
                "data": {
                    "class": str,
                    "optional": False
                },
            })
        )

    @classmethod
    def clearDeviceMetricsOverride(cls):
        """Clears the overriden device metrics.
        """
        return (
            cls.build_send_payload("clearDeviceMetricsOverride", {
            }),
            None
        )

    @classmethod
    def clearDeviceOrientationOverride(cls):
        """Clears the overridden Device Orientation.
        """
        return (
            cls.build_send_payload("clearDeviceOrientationOverride", {
            }),
            None
        )

    @classmethod
    def clearGeolocationOverride(cls):
        """Clears the overriden Geolocation Position and Error.
        """
        return (
            cls.build_send_payload("clearGeolocationOverride", {
            }),
            None
        )

    @classmethod
    def createIsolatedWorld(cls,
                            frameId: Union['FrameId'],
                            worldName: Optional['str'] = None,
                            grantUniveralAccess: Optional['bool'] = None,
                            ):
        """Creates an isolated world for the given frame.
        :param frameId: Id of the frame in which the isolated world should be created.
        :type frameId: FrameId
        :param worldName: An optional name which is reported in the Execution Context.
        :type worldName: str
        :param grantUniveralAccess: Whether or not universal access should be granted to the isolated world. This is a powerful
option, use with caution.
        :type grantUniveralAccess: bool
        """
        return (
            cls.build_send_payload("createIsolatedWorld", {
                "frameId": frameId,
                "worldName": worldName,
                "grantUniveralAccess": grantUniveralAccess,
            }),
            cls.convert_payload({
                "executionContextId": {
                    "class": Runtime.ExecutionContextId,
                    "optional": False
                },
            })
        )

    @classmethod
    def deleteCookie(cls,
                     cookieName: Union['str'],
                     url: Union['str'],
                     ):
        """Deletes browser cookie with given name, domain and path.
        :param cookieName: Name of the cookie to remove.
        :type cookieName: str
        :param url: URL to match cooke domain and path.
        :type url: str
        """
        return (
            cls.build_send_payload("deleteCookie", {
                "cookieName": cookieName,
                "url": url,
            }),
            None
        )

    @classmethod
    def disable(cls):
        """Disables page domain notifications.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables page domain notifications.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def getAppManifest(cls):
        """
        """
        return (
            cls.build_send_payload("getAppManifest", {
            }),
            cls.convert_payload({
                "url": {
                    "class": str,
                    "optional": False
                },
                "errors": {
                    "class": [AppManifestError],
                    "optional": False
                },
                "data": {
                    "class": str,
                    "optional": True
                },
            })
        )

    @classmethod
    def getCookies(cls):
        """Returns all browser cookies. Depending on the backend support, will return detailed cookie
information in the `cookies` field.
        """
        return (
            cls.build_send_payload("getCookies", {
            }),
            cls.convert_payload({
                "cookies": {
                    "class": [Network.Cookie],
                    "optional": False
                },
            })
        )

    @classmethod
    def getFrameTree(cls):
        """Returns present frame tree structure.
        """
        return (
            cls.build_send_payload("getFrameTree", {
            }),
            cls.convert_payload({
                "frameTree": {
                    "class": FrameTree,
                    "optional": False
                },
            })
        )

    @classmethod
    def getLayoutMetrics(cls):
        """Returns metrics relating to the layouting of the page, such as viewport bounds/scale.
        """
        return (
            cls.build_send_payload("getLayoutMetrics", {
            }),
            cls.convert_payload({
                "layoutViewport": {
                    "class": LayoutViewport,
                    "optional": False
                },
                "visualViewport": {
                    "class": VisualViewport,
                    "optional": False
                },
                "contentSize": {
                    "class": DOM.Rect,
                    "optional": False
                },
            })
        )

    @classmethod
    def getNavigationHistory(cls):
        """Returns navigation history for the current page.
        """
        return (
            cls.build_send_payload("getNavigationHistory", {
            }),
            cls.convert_payload({
                "currentIndex": {
                    "class": int,
                    "optional": False
                },
                "entries": {
                    "class": [NavigationEntry],
                    "optional": False
                },
            })
        )

    @classmethod
    def getResourceContent(cls,
                           frameId: Union['FrameId'],
                           url: Union['str'],
                           ):
        """Returns content of the given resource.
        :param frameId: Frame id to get resource for.
        :type frameId: FrameId
        :param url: URL of the resource to get content for.
        :type url: str
        """
        return (
            cls.build_send_payload("getResourceContent", {
                "frameId": frameId,
                "url": url,
            }),
            cls.convert_payload({
                "content": {
                    "class": str,
                    "optional": False
                },
                "base64Encoded": {
                    "class": bool,
                    "optional": False
                },
            })
        )

    @classmethod
    def getResourceTree(cls):
        """Returns present frame / resource tree structure.
        """
        return (
            cls.build_send_payload("getResourceTree", {
            }),
            cls.convert_payload({
                "frameTree": {
                    "class": FrameResourceTree,
                    "optional": False
                },
            })
        )

    @classmethod
    def handleJavaScriptDialog(cls,
                               accept: Union['bool'],
                               promptText: Optional['str'] = None,
                               ):
        """Accepts or dismisses a JavaScript initiated dialog (alert, confirm, prompt, or onbeforeunload).
        :param accept: Whether to accept or dismiss the dialog.
        :type accept: bool
        :param promptText: The text to enter into the dialog prompt before accepting. Used only if this is a prompt
dialog.
        :type promptText: str
        """
        return (
            cls.build_send_payload("handleJavaScriptDialog", {
                "accept": accept,
                "promptText": promptText,
            }),
            None
        )

    @classmethod
    def navigate(cls,
                 url: Union['str'],
                 referrer: Optional['str'] = None,
                 transitionType: Optional['TransitionType'] = None,
                 frameId: Optional['FrameId'] = None,
                 ):
        """Navigates current page to the given URL.
        :param url: URL to navigate the page to.
        :type url: str
        :param referrer: Referrer URL.
        :type referrer: str
        :param transitionType: Intended transition type.
        :type transitionType: TransitionType
        :param frameId: Frame id to navigate, if not specified navigates the top frame.
        :type frameId: FrameId
        """
        return (
            cls.build_send_payload("navigate", {
                "url": url,
                "referrer": referrer,
                "transitionType": transitionType,
                "frameId": frameId,
            }),
            cls.convert_payload({
                "frameId": {
                    "class": FrameId,
                    "optional": False
                },
                "loaderId": {
                    "class": Network.LoaderId,
                    "optional": True
                },
                "errorText": {
                    "class": str,
                    "optional": True
                },
            })
        )

    @classmethod
    def navigateToHistoryEntry(cls,
                               entryId: Union['int'],
                               ):
        """Navigates current page to the given history entry.
        :param entryId: Unique id of the entry to navigate to.
        :type entryId: int
        """
        return (
            cls.build_send_payload("navigateToHistoryEntry", {
                "entryId": entryId,
            }),
            None
        )

    @classmethod
    def printToPDF(cls,
                   landscape: Optional['bool'] = None,
                   displayHeaderFooter: Optional['bool'] = None,
                   printBackground: Optional['bool'] = None,
                   scale: Optional['float'] = None,
                   paperWidth: Optional['float'] = None,
                   paperHeight: Optional['float'] = None,
                   marginTop: Optional['float'] = None,
                   marginBottom: Optional['float'] = None,
                   marginLeft: Optional['float'] = None,
                   marginRight: Optional['float'] = None,
                   pageRanges: Optional['str'] = None,
                   ignoreInvalidPageRanges: Optional['bool'] = None,
                   headerTemplate: Optional['str'] = None,
                   footerTemplate: Optional['str'] = None,
                   preferCSSPageSize: Optional['bool'] = None,
                   ):
        """Print page as PDF.
        :param landscape: Paper orientation. Defaults to false.
        :type landscape: bool
        :param displayHeaderFooter: Display header and footer. Defaults to false.
        :type displayHeaderFooter: bool
        :param printBackground: Print background graphics. Defaults to false.
        :type printBackground: bool
        :param scale: Scale of the webpage rendering. Defaults to 1.
        :type scale: float
        :param paperWidth: Paper width in inches. Defaults to 8.5 inches.
        :type paperWidth: float
        :param paperHeight: Paper height in inches. Defaults to 11 inches.
        :type paperHeight: float
        :param marginTop: Top margin in inches. Defaults to 1cm (~0.4 inches).
        :type marginTop: float
        :param marginBottom: Bottom margin in inches. Defaults to 1cm (~0.4 inches).
        :type marginBottom: float
        :param marginLeft: Left margin in inches. Defaults to 1cm (~0.4 inches).
        :type marginLeft: float
        :param marginRight: Right margin in inches. Defaults to 1cm (~0.4 inches).
        :type marginRight: float
        :param pageRanges: Paper ranges to print, e.g., '1-5, 8, 11-13'. Defaults to the empty string, which means
print all pages.
        :type pageRanges: str
        :param ignoreInvalidPageRanges: Whether to silently ignore invalid but successfully parsed page ranges, such as '3-2'.
Defaults to false.
        :type ignoreInvalidPageRanges: bool
        :param headerTemplate: HTML template for the print header. Should be valid HTML markup with following
classes used to inject printing values into them:
- `date`: formatted print date
- `title`: document title
- `url`: document location
- `pageNumber`: current page number
- `totalPages`: total pages in the document

For example, `<span class=title></span>` would generate span containing the title.
        :type headerTemplate: str
        :param footerTemplate: HTML template for the print footer. Should use the same format as the `headerTemplate`.
        :type footerTemplate: str
        :param preferCSSPageSize: Whether or not to prefer page size as defined by css. Defaults to false,
in which case the content will be scaled to fit the paper size.
        :type preferCSSPageSize: bool
        """
        return (
            cls.build_send_payload("printToPDF", {
                "landscape": landscape,
                "displayHeaderFooter": displayHeaderFooter,
                "printBackground": printBackground,
                "scale": scale,
                "paperWidth": paperWidth,
                "paperHeight": paperHeight,
                "marginTop": marginTop,
                "marginBottom": marginBottom,
                "marginLeft": marginLeft,
                "marginRight": marginRight,
                "pageRanges": pageRanges,
                "ignoreInvalidPageRanges": ignoreInvalidPageRanges,
                "headerTemplate": headerTemplate,
                "footerTemplate": footerTemplate,
                "preferCSSPageSize": preferCSSPageSize,
            }),
            cls.convert_payload({
                "data": {
                    "class": str,
                    "optional": False
                },
            })
        )

    @classmethod
    def reload(cls,
               ignoreCache: Optional['bool'] = None,
               scriptToEvaluateOnLoad: Optional['str'] = None,
               ):
        """Reloads given page optionally ignoring the cache.
        :param ignoreCache: If true, browser cache is ignored (as if the user pressed Shift+refresh).
        :type ignoreCache: bool
        :param scriptToEvaluateOnLoad: If set, the script will be injected into all frames of the inspected page after reload.
Argument will be ignored if reloading dataURL origin.
        :type scriptToEvaluateOnLoad: str
        """
        return (
            cls.build_send_payload("reload", {
                "ignoreCache": ignoreCache,
                "scriptToEvaluateOnLoad": scriptToEvaluateOnLoad,
            }),
            None
        )

    @classmethod
    def removeScriptToEvaluateOnLoad(cls,
                                     identifier: Union['ScriptIdentifier'],
                                     ):
        """Deprecated, please use removeScriptToEvaluateOnNewDocument instead.
        :param identifier: 
        :type identifier: ScriptIdentifier
        """
        return (
            cls.build_send_payload("removeScriptToEvaluateOnLoad", {
                "identifier": identifier,
            }),
            None
        )

    @classmethod
    def removeScriptToEvaluateOnNewDocument(cls,
                                            identifier: Union['ScriptIdentifier'],
                                            ):
        """Removes given script from the list.
        :param identifier: 
        :type identifier: ScriptIdentifier
        """
        return (
            cls.build_send_payload("removeScriptToEvaluateOnNewDocument", {
                "identifier": identifier,
            }),
            None
        )

    @classmethod
    def requestAppBanner(cls):
        """
        """
        return (
            cls.build_send_payload("requestAppBanner", {
            }),
            None
        )

    @classmethod
    def screencastFrameAck(cls,
                           sessionId: Union['int'],
                           ):
        """Acknowledges that a screencast frame has been received by the frontend.
        :param sessionId: Frame number.
        :type sessionId: int
        """
        return (
            cls.build_send_payload("screencastFrameAck", {
                "sessionId": sessionId,
            }),
            None
        )

    @classmethod
    def searchInResource(cls,
                         frameId: Union['FrameId'],
                         url: Union['str'],
                         query: Union['str'],
                         caseSensitive: Optional['bool'] = None,
                         isRegex: Optional['bool'] = None,
                         ):
        """Searches for given string in resource content.
        :param frameId: Frame id for resource to search in.
        :type frameId: FrameId
        :param url: URL of the resource to search in.
        :type url: str
        :param query: String to search for.
        :type query: str
        :param caseSensitive: If true, search is case sensitive.
        :type caseSensitive: bool
        :param isRegex: If true, treats string parameter as regex.
        :type isRegex: bool
        """
        return (
            cls.build_send_payload("searchInResource", {
                "frameId": frameId,
                "url": url,
                "query": query,
                "caseSensitive": caseSensitive,
                "isRegex": isRegex,
            }),
            cls.convert_payload({
                "result": {
                    "class": [Debugger.SearchMatch],
                    "optional": False
                },
            })
        )

    @classmethod
    def setAdBlockingEnabled(cls,
                             enabled: Union['bool'],
                             ):
        """Enable Chrome's experimental ad filter on all sites.
        :param enabled: Whether to block ads.
        :type enabled: bool
        """
        return (
            cls.build_send_payload("setAdBlockingEnabled", {
                "enabled": enabled,
            }),
            None
        )

    @classmethod
    def setBypassCSP(cls,
                     enabled: Union['bool'],
                     ):
        """Enable page Content Security Policy by-passing.
        :param enabled: Whether to bypass page CSP.
        :type enabled: bool
        """
        return (
            cls.build_send_payload("setBypassCSP", {
                "enabled": enabled,
            }),
            None
        )

    @classmethod
    def setDeviceMetricsOverride(cls,
                                 width: Union['int'],
                                 height: Union['int'],
                                 deviceScaleFactor: Union['float'],
                                 mobile: Union['bool'],
                                 scale: Optional['float'] = None,
                                 screenWidth: Optional['int'] = None,
                                 screenHeight: Optional['int'] = None,
                                 positionX: Optional['int'] = None,
                                 positionY: Optional['int'] = None,
                                 dontSetVisibleSize: Optional['bool'] = None,
                                 screenOrientation: Optional['Emulation.ScreenOrientation'] = None,
                                 viewport: Optional['Viewport'] = None,
                                 ):
        """Overrides the values of device screen dimensions (window.screen.width, window.screen.height,
window.innerWidth, window.innerHeight, and "device-width"/"device-height"-related CSS media
query results).
        :param width: Overriding width value in pixels (minimum 0, maximum 10000000). 0 disables the override.
        :type width: int
        :param height: Overriding height value in pixels (minimum 0, maximum 10000000). 0 disables the override.
        :type height: int
        :param deviceScaleFactor: Overriding device scale factor value. 0 disables the override.
        :type deviceScaleFactor: float
        :param mobile: Whether to emulate mobile device. This includes viewport meta tag, overlay scrollbars, text
autosizing and more.
        :type mobile: bool
        :param scale: Scale to apply to resulting view image.
        :type scale: float
        :param screenWidth: Overriding screen width value in pixels (minimum 0, maximum 10000000).
        :type screenWidth: int
        :param screenHeight: Overriding screen height value in pixels (minimum 0, maximum 10000000).
        :type screenHeight: int
        :param positionX: Overriding view X position on screen in pixels (minimum 0, maximum 10000000).
        :type positionX: int
        :param positionY: Overriding view Y position on screen in pixels (minimum 0, maximum 10000000).
        :type positionY: int
        :param dontSetVisibleSize: Do not set visible view size, rely upon explicit setVisibleSize call.
        :type dontSetVisibleSize: bool
        :param screenOrientation: Screen orientation override.
        :type screenOrientation: Emulation.ScreenOrientation
        :param viewport: The viewport dimensions and scale. If not set, the override is cleared.
        :type viewport: Viewport
        """
        return (
            cls.build_send_payload("setDeviceMetricsOverride", {
                "width": width,
                "height": height,
                "deviceScaleFactor": deviceScaleFactor,
                "mobile": mobile,
                "scale": scale,
                "screenWidth": screenWidth,
                "screenHeight": screenHeight,
                "positionX": positionX,
                "positionY": positionY,
                "dontSetVisibleSize": dontSetVisibleSize,
                "screenOrientation": screenOrientation,
                "viewport": viewport,
            }),
            None
        )

    @classmethod
    def setDeviceOrientationOverride(cls,
                                     alpha: Union['float'],
                                     beta: Union['float'],
                                     gamma: Union['float'],
                                     ):
        """Overrides the Device Orientation.
        :param alpha: Mock alpha
        :type alpha: float
        :param beta: Mock beta
        :type beta: float
        :param gamma: Mock gamma
        :type gamma: float
        """
        return (
            cls.build_send_payload("setDeviceOrientationOverride", {
                "alpha": alpha,
                "beta": beta,
                "gamma": gamma,
            }),
            None
        )

    @classmethod
    def setFontFamilies(cls,
                        fontFamilies: Union['FontFamilies'],
                        ):
        """Set generic font families.
        :param fontFamilies: Specifies font families to set. If a font family is not specified, it won't be changed.
        :type fontFamilies: FontFamilies
        """
        return (
            cls.build_send_payload("setFontFamilies", {
                "fontFamilies": fontFamilies,
            }),
            None
        )

    @classmethod
    def setFontSizes(cls,
                     fontSizes: Union['FontSizes'],
                     ):
        """Set default font sizes.
        :param fontSizes: Specifies font sizes to set. If a font size is not specified, it won't be changed.
        :type fontSizes: FontSizes
        """
        return (
            cls.build_send_payload("setFontSizes", {
                "fontSizes": fontSizes,
            }),
            None
        )

    @classmethod
    def setDocumentContent(cls,
                           frameId: Union['FrameId'],
                           html: Union['str'],
                           ):
        """Sets given markup as the document's HTML.
        :param frameId: Frame id to set HTML for.
        :type frameId: FrameId
        :param html: HTML content to set.
        :type html: str
        """
        return (
            cls.build_send_payload("setDocumentContent", {
                "frameId": frameId,
                "html": html,
            }),
            None
        )

    @classmethod
    def setDownloadBehavior(cls,
                            behavior: Union['str'],
                            downloadPath: Optional['str'] = None,
                            ):
        """Set the behavior when downloading a file.
        :param behavior: Whether to allow all or deny all download requests, or use default Chrome behavior if
available (otherwise deny).
        :type behavior: str
        :param downloadPath: The default path to save downloaded files to. This is requred if behavior is set to 'allow'
        :type downloadPath: str
        """
        return (
            cls.build_send_payload("setDownloadBehavior", {
                "behavior": behavior,
                "downloadPath": downloadPath,
            }),
            None
        )

    @classmethod
    def setGeolocationOverride(cls,
                               latitude: Optional['float'] = None,
                               longitude: Optional['float'] = None,
                               accuracy: Optional['float'] = None,
                               ):
        """Overrides the Geolocation Position or Error. Omitting any of the parameters emulates position
unavailable.
        :param latitude: Mock latitude
        :type latitude: float
        :param longitude: Mock longitude
        :type longitude: float
        :param accuracy: Mock accuracy
        :type accuracy: float
        """
        return (
            cls.build_send_payload("setGeolocationOverride", {
                "latitude": latitude,
                "longitude": longitude,
                "accuracy": accuracy,
            }),
            None
        )

    @classmethod
    def setLifecycleEventsEnabled(cls,
                                  enabled: Union['bool'],
                                  ):
        """Controls whether page will emit lifecycle events.
        :param enabled: If true, starts emitting lifecycle events.
        :type enabled: bool
        """
        return (
            cls.build_send_payload("setLifecycleEventsEnabled", {
                "enabled": enabled,
            }),
            None
        )

    @classmethod
    def setTouchEmulationEnabled(cls,
                                 enabled: Union['bool'],
                                 configuration: Optional['str'] = None,
                                 ):
        """Toggles mouse event-based touch event emulation.
        :param enabled: Whether the touch event emulation should be enabled.
        :type enabled: bool
        :param configuration: Touch/gesture events configuration. Default: current platform.
        :type configuration: str
        """
        return (
            cls.build_send_payload("setTouchEmulationEnabled", {
                "enabled": enabled,
                "configuration": configuration,
            }),
            None
        )

    @classmethod
    def startScreencast(cls,
                        format: Optional['str'] = None,
                        quality: Optional['int'] = None,
                        maxWidth: Optional['int'] = None,
                        maxHeight: Optional['int'] = None,
                        everyNthFrame: Optional['int'] = None,
                        ):
        """Starts sending each frame using the `screencastFrame` event.
        :param format: Image compression format.
        :type format: str
        :param quality: Compression quality from range [0..100].
        :type quality: int
        :param maxWidth: Maximum screenshot width.
        :type maxWidth: int
        :param maxHeight: Maximum screenshot height.
        :type maxHeight: int
        :param everyNthFrame: Send every n-th frame.
        :type everyNthFrame: int
        """
        return (
            cls.build_send_payload("startScreencast", {
                "format": format,
                "quality": quality,
                "maxWidth": maxWidth,
                "maxHeight": maxHeight,
                "everyNthFrame": everyNthFrame,
            }),
            None
        )

    @classmethod
    def stopLoading(cls):
        """Force the page stop all navigations and pending resource fetches.
        """
        return (
            cls.build_send_payload("stopLoading", {
            }),
            None
        )

    @classmethod
    def crash(cls):
        """Crashes renderer on the IO thread, generates minidumps.
        """
        return (
            cls.build_send_payload("crash", {
            }),
            None
        )

    @classmethod
    def close(cls):
        """Tries to close page, running its beforeunload hooks, if any.
        """
        return (
            cls.build_send_payload("close", {
            }),
            None
        )

    @classmethod
    def setWebLifecycleState(cls,
                             state: Union['str'],
                             ):
        """Tries to update the web lifecycle state of the page.
It will transition the page to the given state according to:
https://github.com/WICG/web-lifecycle/
        :param state: Target lifecycle state
        :type state: str
        """
        return (
            cls.build_send_payload("setWebLifecycleState", {
                "state": state,
            }),
            None
        )

    @classmethod
    def stopScreencast(cls):
        """Stops sending each frame in the `screencastFrame`.
        """
        return (
            cls.build_send_payload("stopScreencast", {
            }),
            None
        )

    @classmethod
    def setProduceCompilationCache(cls,
                                   enabled: Union['bool'],
                                   ):
        """Forces compilation cache to be generated for every subresource script.
        :param enabled: 
        :type enabled: bool
        """
        return (
            cls.build_send_payload("setProduceCompilationCache", {
                "enabled": enabled,
            }),
            None
        )

    @classmethod
    def addCompilationCache(cls,
                            url: Union['str'],
                            data: Union['str'],
                            ):
        """Seeds compilation cache for given url. Compilation cache does not survive
cross-process navigation.
        :param url: 
        :type url: str
        :param data: Base64-encoded data
        :type data: str
        """
        return (
            cls.build_send_payload("addCompilationCache", {
                "url": url,
                "data": data,
            }),
            None
        )

    @classmethod
    def clearCompilationCache(cls):
        """Clears seeded compilation cache.
        """
        return (
            cls.build_send_payload("clearCompilationCache", {
            }),
            None
        )



class DomContentEventFiredEvent(BaseEvent):

    js_name = 'Page.domContentEventFired'
    hashable = []
    is_hashable = False

    def __init__(self,
                 timestamp: Union['Network.MonotonicTime', dict],
                 ):
        if isinstance(timestamp, dict):
            timestamp = Network.MonotonicTime(**timestamp)
        self.timestamp = timestamp

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class FrameAttachedEvent(BaseEvent):

    js_name = 'Page.frameAttached'
    hashable = ['parentFrameId', 'frameId']
    is_hashable = True

    def __init__(self,
                 frameId: Union['FrameId', dict],
                 parentFrameId: Union['FrameId', dict],
                 stack: Union['Runtime.StackTrace', dict, None] = None,
                 ):
        if isinstance(frameId, dict):
            frameId = FrameId(**frameId)
        self.frameId = frameId
        if isinstance(parentFrameId, dict):
            parentFrameId = FrameId(**parentFrameId)
        self.parentFrameId = parentFrameId
        if isinstance(stack, dict):
            stack = Runtime.StackTrace(**stack)
        self.stack = stack

    @classmethod
    def build_hash(cls, parentFrameId, frameId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class FrameClearedScheduledNavigationEvent(BaseEvent):

    js_name = 'Page.frameClearedScheduledNavigation'
    hashable = ['frameId']
    is_hashable = True

    def __init__(self,
                 frameId: Union['FrameId', dict],
                 ):
        if isinstance(frameId, dict):
            frameId = FrameId(**frameId)
        self.frameId = frameId

    @classmethod
    def build_hash(cls, frameId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class FrameDetachedEvent(BaseEvent):

    js_name = 'Page.frameDetached'
    hashable = ['frameId']
    is_hashable = True

    def __init__(self,
                 frameId: Union['FrameId', dict],
                 ):
        if isinstance(frameId, dict):
            frameId = FrameId(**frameId)
        self.frameId = frameId

    @classmethod
    def build_hash(cls, frameId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class FrameNavigatedEvent(BaseEvent):

    js_name = 'Page.frameNavigated'
    hashable = ['frameId']
    is_hashable = True

    def __init__(self,
                 frame: Union['Frame', dict],
                 ):
        if isinstance(frame, dict):
            frame = Frame(**frame)
        self.frame = frame

    @classmethod
    def build_hash(cls, frameId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class FrameResizedEvent(BaseEvent):

    js_name = 'Page.frameResized'
    hashable = []
    is_hashable = False

    def __init__(self):
        pass

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class FrameScheduledNavigationEvent(BaseEvent):

    js_name = 'Page.frameScheduledNavigation'
    hashable = ['frameId']
    is_hashable = True

    def __init__(self,
                 frameId: Union['FrameId', dict],
                 delay: Union['float', dict],
                 reason: Union['str', dict],
                 url: Union['str', dict],
                 ):
        if isinstance(frameId, dict):
            frameId = FrameId(**frameId)
        self.frameId = frameId
        if isinstance(delay, dict):
            delay = float(**delay)
        self.delay = delay
        if isinstance(reason, dict):
            reason = str(**reason)
        self.reason = reason
        if isinstance(url, dict):
            url = str(**url)
        self.url = url

    @classmethod
    def build_hash(cls, frameId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class FrameStartedLoadingEvent(BaseEvent):

    js_name = 'Page.frameStartedLoading'
    hashable = ['frameId']
    is_hashable = True

    def __init__(self,
                 frameId: Union['FrameId', dict],
                 ):
        if isinstance(frameId, dict):
            frameId = FrameId(**frameId)
        self.frameId = frameId

    @classmethod
    def build_hash(cls, frameId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class FrameStoppedLoadingEvent(BaseEvent):

    js_name = 'Page.frameStoppedLoading'
    hashable = ['frameId']
    is_hashable = True

    def __init__(self,
                 frameId: Union['FrameId', dict],
                 ):
        if isinstance(frameId, dict):
            frameId = FrameId(**frameId)
        self.frameId = frameId

    @classmethod
    def build_hash(cls, frameId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class InterstitialHiddenEvent(BaseEvent):

    js_name = 'Page.interstitialHidden'
    hashable = []
    is_hashable = False

    def __init__(self):
        pass

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class InterstitialShownEvent(BaseEvent):

    js_name = 'Page.interstitialShown'
    hashable = []
    is_hashable = False

    def __init__(self):
        pass

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class JavascriptDialogClosedEvent(BaseEvent):

    js_name = 'Page.javascriptDialogClosed'
    hashable = []
    is_hashable = False

    def __init__(self,
                 result: Union['bool', dict],
                 userInput: Union['str', dict],
                 ):
        if isinstance(result, dict):
            result = bool(**result)
        self.result = result
        if isinstance(userInput, dict):
            userInput = str(**userInput)
        self.userInput = userInput

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class JavascriptDialogOpeningEvent(BaseEvent):

    js_name = 'Page.javascriptDialogOpening'
    hashable = []
    is_hashable = False

    def __init__(self,
                 url: Union['str', dict],
                 message: Union['str', dict],
                 type: Union['DialogType', dict],
                 hasBrowserHandler: Union['bool', dict],
                 defaultPrompt: Union['str', dict, None] = None,
                 ):
        if isinstance(url, dict):
            url = str(**url)
        self.url = url
        if isinstance(message, dict):
            message = str(**message)
        self.message = message
        if isinstance(type, dict):
            type = DialogType(**type)
        self.type = type
        if isinstance(hasBrowserHandler, dict):
            hasBrowserHandler = bool(**hasBrowserHandler)
        self.hasBrowserHandler = hasBrowserHandler
        if isinstance(defaultPrompt, dict):
            defaultPrompt = str(**defaultPrompt)
        self.defaultPrompt = defaultPrompt

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class LifecycleEventEvent(BaseEvent):

    js_name = 'Page.lifecycleEvent'
    hashable = ['loaderId', 'frameId']
    is_hashable = True

    def __init__(self,
                 frameId: Union['FrameId', dict],
                 loaderId: Union['Network.LoaderId', dict],
                 name: Union['str', dict],
                 timestamp: Union['Network.MonotonicTime', dict],
                 ):
        if isinstance(frameId, dict):
            frameId = FrameId(**frameId)
        self.frameId = frameId
        if isinstance(loaderId, dict):
            loaderId = Network.LoaderId(**loaderId)
        self.loaderId = loaderId
        if isinstance(name, dict):
            name = str(**name)
        self.name = name
        if isinstance(timestamp, dict):
            timestamp = Network.MonotonicTime(**timestamp)
        self.timestamp = timestamp

    @classmethod
    def build_hash(cls, loaderId, frameId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class LoadEventFiredEvent(BaseEvent):

    js_name = 'Page.loadEventFired'
    hashable = []
    is_hashable = False

    def __init__(self,
                 timestamp: Union['Network.MonotonicTime', dict],
                 ):
        if isinstance(timestamp, dict):
            timestamp = Network.MonotonicTime(**timestamp)
        self.timestamp = timestamp

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class NavigatedWithinDocumentEvent(BaseEvent):

    js_name = 'Page.navigatedWithinDocument'
    hashable = ['frameId']
    is_hashable = True

    def __init__(self,
                 frameId: Union['FrameId', dict],
                 url: Union['str', dict],
                 ):
        if isinstance(frameId, dict):
            frameId = FrameId(**frameId)
        self.frameId = frameId
        if isinstance(url, dict):
            url = str(**url)
        self.url = url

    @classmethod
    def build_hash(cls, frameId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ScreencastFrameEvent(BaseEvent):

    js_name = 'Page.screencastFrame'
    hashable = ['sessionId']
    is_hashable = True

    def __init__(self,
                 data: Union['str', dict],
                 metadata: Union['ScreencastFrameMetadata', dict],
                 sessionId: Union['int', dict],
                 ):
        if isinstance(data, dict):
            data = str(**data)
        self.data = data
        if isinstance(metadata, dict):
            metadata = ScreencastFrameMetadata(**metadata)
        self.metadata = metadata
        if isinstance(sessionId, dict):
            sessionId = int(**sessionId)
        self.sessionId = sessionId

    @classmethod
    def build_hash(cls, sessionId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ScreencastVisibilityChangedEvent(BaseEvent):

    js_name = 'Page.screencastVisibilityChanged'
    hashable = []
    is_hashable = False

    def __init__(self,
                 visible: Union['bool', dict],
                 ):
        if isinstance(visible, dict):
            visible = bool(**visible)
        self.visible = visible

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class WindowOpenEvent(BaseEvent):

    js_name = 'Page.windowOpen'
    hashable = []
    is_hashable = False

    def __init__(self,
                 url: Union['str', dict],
                 windowName: Union['str', dict],
                 windowFeatures: Union['[]', dict],
                 userGesture: Union['bool', dict],
                 ):
        if isinstance(url, dict):
            url = str(**url)
        self.url = url
        if isinstance(windowName, dict):
            windowName = str(**windowName)
        self.windowName = windowName
        if isinstance(windowFeatures, dict):
            windowFeatures = [](**windowFeatures)
        self.windowFeatures = windowFeatures
        if isinstance(userGesture, dict):
            userGesture = bool(**userGesture)
        self.userGesture = userGesture

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class CompilationCacheProducedEvent(BaseEvent):

    js_name = 'Page.compilationCacheProduced'
    hashable = []
    is_hashable = False

    def __init__(self,
                 url: Union['str', dict],
                 data: Union['str', dict],
                 ):
        if isinstance(url, dict):
            url = str(**url)
        self.url = url
        if isinstance(data, dict):
            data = str(**data)
        self.data = data

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import network as Network

class Audits(PayloadMixin):
    """ Audits domain allows investigation of page violations and possible improvements.
    """
    @classmethod
    def getEncodedResponse(cls,
                           requestId: Union['Network.RequestId'],
                           encoding: Union['str'],
                           quality: Optional['float'] = None,
                           sizeOnly: Optional['bool'] = None,
                           ):
        """Returns the response body and size if it were re-encoded with the specified settings. Only
applies to images.
        :param requestId: Identifier of the network request to get content for.
        :type requestId: Network.RequestId
        :param encoding: The encoding to use.
        :type encoding: str
        :param quality: The quality of the encoding (0-1). (defaults to 1)
        :type quality: float
        :param sizeOnly: Whether to only return the size information (defaults to false).
        :type sizeOnly: bool
        """
        return (
            cls.build_send_payload("getEncodedResponse", {
                "requestId": requestId,
                "encoding": encoding,
                "quality": quality,
                "sizeOnly": sizeOnly,
            }),
            cls.convert_payload({
                "body": {
                    "class": str,
                    "optional": True
                },
                "originalSize": {
                    "class": int,
                    "optional": False
                },
                "encodedSize": {
                    "class": int,
                    "optional": False
                },
            })
        )




import asyncio
import base64
import json
import logging
from typing import Optional

import aiohttp
import websockets
import websockets.protocol

from chromewhip import helpers
from chromewhip.base import SyncAdder
from chromewhip.protocol import page, runtime, target, input, inspector, browser, accessibility

TIMEOUT_S = 25
MAX_PAYLOAD_SIZE_BYTES = 2 ** 23
MAX_PAYLOAD_SIZE_MB = MAX_PAYLOAD_SIZE_BYTES / 1024 ** 2


class ChromewhipException(Exception):
    pass


class TimeoutError(Exception):
    pass


class ProtocolError(ChromewhipException):
    pass


class JSScriptError(ChromewhipException):
    pass


class ChromeTab(metaclass=SyncAdder):

    def __init__(self, title, url, ws_uri, tab_id):
        self.id_ = tab_id
        self._title = title
        self._url = url
        self._ws_uri = ws_uri
        self.target_id = ws_uri.split('/')[-1]
        self._ws: Optional[websockets.WebSocketClientProtocol] = None
        self._message_id = 0
        self._current_task: Optional[asyncio.Task] = None
        self._ack_events = {}
        self._ack_payloads = {}
        self._input_events = {}
        self._trigger_events = {}
        self._event_payloads = {}
        self._recv_task = None
        self._log = logging.getLogger('chromewhip.chrome.ChromeTab')
        self._send_log = logging.getLogger('chromewhip.chrome.ChromeTab.send_handler')
        self._recv_log = logging.getLogger('chromewhip.chrome.ChromeTab.recv_handler')

    @classmethod
    async def create_from_json(cls, json_, host, port):
        ws_url = json_.get('webSocketDebuggerUrl')
        if not ws_url:
            tab_id = json_['id']
            ws_url = 'ws://{}:{}/devtools/page/{}'.format(host,
                                                          port,
                                                          tab_id)
        t = cls(json_['title'], json_['url'],  ws_url, json_['id'])
        await t.connect()
        return t

    async def connect(self):
        self._ws = await websockets.connect(self._ws_uri, max_size=MAX_PAYLOAD_SIZE_BYTES)  # 16MB
        self._recv_task = asyncio.ensure_future(self.recv_handler())
        self._log.info('Connected to Chrome tab %s' % self._ws_uri)

    async def disconnect(self):
        self._log.debug("Disconnecting tab...")
        if self._current_task and not self._current_task.done() and not self._current_task.cancelled():
            self._log.warning('Cancelling current task for websocket')
            self._current_task.cancel()
            await self._current_task
        if self._recv_task:
            self._recv_task.cancel()
            await self._recv_task

    async def recv_handler(self):
        try:
            while True:
                self._recv_log.debug('Waiting for message...')
                result = await self._ws.recv()
                self._recv_log.debug('Received message, processing...')

                if not result:
                    self._recv_log.error('Missing message, may have been a connection timeout...')
                    continue
                result = json.loads(result)

                if not isinstance(result, dict):
                    self._recv_log.error('decoded messages is of type "%s" and = "%s"' % (type(result), result))
                    continue
                if 'id' in result:
                    self._ack_payloads[result['id']] = result
                    ack_event = self._ack_events.get(result['id'])
                    if ack_event is None:
                        self._recv_log.error('Ignoring ack with id %s as no registered recv' % result['id'])
                        continue
                    self._recv_log.debug('Notifying ack event with id=%s' % (result['id']))
                    ack_event.set()

                elif 'method' in result:
                    self._recv_log.debug('Received event message!')
                    event = helpers.json_to_event(result)
                    self._recv_log.debug('Received a "%s" event , storing against hash and name...' % event.js_name)
                    hash_ = event.hash_()
                    self._event_payloads[hash_] = event
                    self._event_payloads[event.js_name] = event

                    # first, check if any requests are waiting upon it
                    input_event = self._input_events.get(event.js_name)
                    if input_event:
                        self._recv_log.debug('input exists for event name "%s", alerting...' % event.js_name)
                        input_event.set()

                    trigger_event = self._trigger_events.get(hash_)
                    if trigger_event:
                        self._recv_log.debug('trigger exists for hash "%s", alerting...' % hash_)
                        trigger_event.set()
                else:
                    # TODO: deal with invalid state
                    self._recv_log.info('Invalid message %s, what do i do now?' % result)

        except asyncio.CancelledError:
            await self._ws.close()

    @staticmethod
    async def validator(result: dict, types: dict):
        for k, v in result.items():
            try:
                type_ = types[k]
            except KeyError:
                raise KeyError('%s not in expected payload of %s' % (k, types))
            if not isinstance(v, type_):
                raise ValueError('%s is not expected type %s, instead is %s' % (v, type_, type(v)))
        return result

    async def _send(self, request, recv_validator=None, input_event_cls=None, trigger_event_cls=None):
        """
        TODO:
        * clean up of stale events in payloads and asyncio event stores
        :param request:
        :param recv_validator:
        :param input_event_cls:
        :param trigger_event_cls:
        :return:
        """
        self._message_id += 1
        request['id'] = self._message_id

        ack_event = asyncio.Event()
        self._ack_events[self._message_id] = ack_event

        if input_event_cls:
            if not input_event_cls.is_hashable:
                raise ValueError('Input event class "%s" as not hashable' % input_event_cls.__name__)
            # we can already register the input event before sending command
            input_event = asyncio.Event()
            self._input_events[input_event_cls.js_name] = input_event

        if trigger_event_cls:
            if not trigger_event_cls.is_hashable:
                raise ValueError('Trigger event type "%s" as not hashable' % trigger_event_cls.__name__)

        result = {'ack': None, 'event': None}

        try:
            msg = json.dumps(request, cls=helpers.ChromewhipJSONEncoder)
            self._send_log.info('Sending command = %s' % msg)
            self._current_task = asyncio.ensure_future(self._ws.send(msg))
            await asyncio.wait_for(self._current_task, timeout=TIMEOUT_S)  # send

            self._send_log.debug('Waiting for ack event set for id=%s' % request['id'])
            await asyncio.wait_for(ack_event.wait(), timeout=TIMEOUT_S)  # recv
            self._send_log.debug('Received ack event set for id=%s' % request['id'])

            ack_payload = self._ack_payloads.get(request['id'])

            if not ack_payload:
                self._send_log.error('Notified but no payload available for id=%s!' % request['id'])
                return result

            # check for errors
            error = ack_payload.get('error')

            if error:
                msg = '%s, code %s for id=%s' % (error.get('message', 'Unknown error'), error['code'], request['id'])
                self._send_log.error(msg)
                raise ProtocolError(msg)

            if recv_validator:
                self._send_log.debug('Validating recv payload for id=%s...' % request['id'])
                ack_result = recv_validator(ack_payload['result'])
                self._send_log.debug('Successful recv validation for id=%s...' % request['id'])
                ack_payload['result'] = ack_result
            else:
                ack_result = ack_payload['result']

            result['ack'] = ack_payload

            if input_event_cls:
                hash_ = input_event_cls.js_name
                # use latest payload as key is not unique within a single session
                event = self._event_payloads.get(hash_)
                hash_input_dict = {}
                if not event:
                    self._send_log.debug('Waiting for event with hash "%s"...' % hash_)
                    await asyncio.wait_for(input_event.wait(), timeout=TIMEOUT_S)  # recv
                    event = self._event_payloads.get(hash_)

                params = event.hash_().split(':')[-1].split(',')
                for p in params:
                    kv = p.split('=')
                    hash_input_dict[kv[0]] = kv[1]
            else:
                hash_input_dict = ack_result

            if trigger_event_cls:
                try:
                    # TODO: put in a `strict` flag so that we can catch differences between the protocol spec and the
                    # underlying implementation.
                    cleaned_hash_input_dict = {k: v for k, v in hash_input_dict.items() if k in trigger_event_cls.hashable}
                    hash_ = trigger_event_cls.build_hash(**cleaned_hash_input_dict)
                except TypeError:
                    raise TypeError(f'Event "{trigger_event_cls.js_name}" hash cannot be built with "{hash_input_dict}"')
                event = self._event_payloads.get(hash_)
                if not event:
                    self._send_log.debug('Waiting for event with hash "%s"...' % hash_)
                    trigger_event = asyncio.Event()
                    self._trigger_events[hash_] = trigger_event
                    await asyncio.wait_for(trigger_event.wait(), timeout=TIMEOUT_S)  # recv
                    event = self._event_payloads.get(hash_)
                result['event'] = event

            self._send_log.info('Successfully sent command = %s' % msg)
            return result
        except asyncio.TimeoutError:
            method = request['method']
            id_ = request['id']
            self._send_log.error(msg)
            if self._ws.state != websockets.protocol.OPEN:
                close_code = self._ws.close_code
                if close_code == 1002:
                    raise ProtocolError('Websocket protocol error occured for "%s" with id=%s' % (method, id_))
                elif close_code == 1006:
                    raise ProtocolError('Incomplete read error occured for "%s" with id=%s' % (method, id_))
                elif close_code == 1007:
                    raise ProtocolError('Unicode decode error occured for "%s" with id=%s' % (method, id_))
                elif close_code == 1009:
                    raise ProtocolError('Recv\'d payload exceeded %sMB for "%s" with id=%s, consider increasing this limit' % (MAX_PAYLOAD_SIZE_MB, method, id_))
            raise TimeoutError('Unknown cause for timeout to occurs for "%s" with id=%s' % (method, id_))

    async def new_message_handler(self, request):
        request['id'] = self._message_id
        await self._ws.send(json.dumps(request))
        return await self._ws.recv()

    @property
    def title(self):
        return self._title

    @property
    def url(self):
        return self._url

    @property
    def ws_uri(self):
        return self._ws_uri

    async def enable_page_events(self):
        return await self._send(*page.Page.enable())

    async def send_command(self, command, input_event_type=None, await_on_event_type=None):
        return await self._send(*command, input_event_cls=input_event_type, trigger_event_cls=await_on_event_type)

    async def html(self):
        result = await self.evaluate('document.documentElement.outerHTML')
        value = result['ack']['result']['result'].value
        return value.encode('utf-8')

    async def screenshot(self):
        result = await self.send_command(page.Page.captureScreenshot(format='png', fromSurface=False))
        base64_data = result['ack']['result']['data']
        return base64.b64decode(base64_data)

    async def go(self, url):
        """
        Navigate the tab to the URL
        """
        return await self.send_command(page.Page.navigate(url),
                                       await_on_event_type=page.FrameStoppedLoadingEvent)

    async def evaluate(self, javascript):
        """
        Evaluate JavaScript on the page
        """
        result = await self.send_command(runtime.Runtime.evaluate(javascript))
        r = result["ack"]["result"]["result"]
        if r.subtype == 'error':
            raise JSScriptError({
                'reason': 'Runtime.evalulate threw an error',
                'error': result["ack"]["result"]["exceptionDetails"].to_dict()
            })
        return result

    def __str__(self):
        return '%s - %s' % (self.title, self.url)

    def __repr__(self):
        return f'ChromeTab("{self.title}", "{self.url}", "{self.ws_uri}, "{self.id_}")'


class Chrome(metaclass=SyncAdder):

    def __init__(self, host='localhost', port=9222):
        self._host = host
        self._port = port
        self._url = 'http://%s:%d' % (self.host, self.port)
        self._tabs = []
        self.is_connected = False
        self._log = logging.getLogger('chromewhip.chrome.Chrome')

    async def connect(self):
        """ Get all open browser tabs that are pages tabs
        """
        if not self.is_connected:
            try:
                await asyncio.wait_for(self.attempt_tab_fetch(), timeout=5)
            except TimeoutError:
                self._log.error('Unable to fetch tabs! Timeout')

    async def attempt_tab_fetch(self):
        async with aiohttp.ClientSession() as session:
            async with session.get(self._url + '/json') as resp:
                tabs = []
                data = await resp.json()
                if not len(data):
                    self._log.warning('Empty data, will attempt to reconnect until able to get pages.')
                for tab in filter(lambda x: x['type'] == 'page', data):
                    t = await ChromeTab.create_from_json(tab, self._host, self._port)
                    tabs.append(t)
                self._tabs = tabs
                self._log.debug("Connected to Chrome! Found {} tabs".format(len(self._tabs)))
        self.is_connected = True


    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def url(self):
        return self._url

    @property
    def tabs(self):
        if not len(self._tabs):
            raise ValueError('Must call connect_s or connect first!')
        return tuple(self._tabs)

    async def create_tab(self):
        async with aiohttp.ClientSession() as session:
            async with session.get(self._url + '/json/new') as resp:
                data = await resp.json()
                t = await ChromeTab.create_from_json(data, self._host, self._port)
                self._tabs.append(t)
        return t

    async def close_tab(self, tab):
        await tab.disconnect()
        async with aiohttp.ClientSession() as session:
            await session.get(self._url + f'/json/close/{tab.id_}')





# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import io as IO

# MemoryDumpConfig: Configuration for memory dump. Used only when "memory-infra" category is enabled.
MemoryDumpConfig = dict

# TraceConfig: 
class TraceConfig(ChromeTypeBase):
    def __init__(self,
                 recordMode: Optional['str'] = None,
                 enableSampling: Optional['bool'] = None,
                 enableSystrace: Optional['bool'] = None,
                 enableArgumentFilter: Optional['bool'] = None,
                 includedCategories: Optional['[]'] = None,
                 excludedCategories: Optional['[]'] = None,
                 syntheticDelays: Optional['[]'] = None,
                 memoryDumpConfig: Optional['MemoryDumpConfig'] = None,
                 ):

        self.recordMode = recordMode
        self.enableSampling = enableSampling
        self.enableSystrace = enableSystrace
        self.enableArgumentFilter = enableArgumentFilter
        self.includedCategories = includedCategories
        self.excludedCategories = excludedCategories
        self.syntheticDelays = syntheticDelays
        self.memoryDumpConfig = memoryDumpConfig


# StreamCompression: Compression type to use for traces returned via streams.
StreamCompression = str

class Tracing(PayloadMixin):
    """ 
    """
    @classmethod
    def end(cls):
        """Stop trace events collection.
        """
        return (
            cls.build_send_payload("end", {
            }),
            None
        )

    @classmethod
    def getCategories(cls):
        """Gets supported tracing categories.
        """
        return (
            cls.build_send_payload("getCategories", {
            }),
            cls.convert_payload({
                "categories": {
                    "class": [],
                    "optional": False
                },
            })
        )

    @classmethod
    def recordClockSyncMarker(cls,
                              syncId: Union['str'],
                              ):
        """Record a clock sync marker in the trace.
        :param syncId: The ID of this clock sync marker
        :type syncId: str
        """
        return (
            cls.build_send_payload("recordClockSyncMarker", {
                "syncId": syncId,
            }),
            None
        )

    @classmethod
    def requestMemoryDump(cls):
        """Request a global memory dump.
        """
        return (
            cls.build_send_payload("requestMemoryDump", {
            }),
            cls.convert_payload({
                "dumpGuid": {
                    "class": str,
                    "optional": False
                },
                "success": {
                    "class": bool,
                    "optional": False
                },
            })
        )

    @classmethod
    def start(cls,
              categories: Optional['str'] = None,
              options: Optional['str'] = None,
              bufferUsageReportingInterval: Optional['float'] = None,
              transferMode: Optional['str'] = None,
              streamCompression: Optional['StreamCompression'] = None,
              traceConfig: Optional['TraceConfig'] = None,
              ):
        """Start trace events collection.
        :param categories: Category/tag filter
        :type categories: str
        :param options: Tracing options
        :type options: str
        :param bufferUsageReportingInterval: If set, the agent will issue bufferUsage events at this interval, specified in milliseconds
        :type bufferUsageReportingInterval: float
        :param transferMode: Whether to report trace events as series of dataCollected events or to save trace to a
stream (defaults to `ReportEvents`).
        :type transferMode: str
        :param streamCompression: Compression format to use. This only applies when using `ReturnAsStream`
transfer mode (defaults to `none`)
        :type streamCompression: StreamCompression
        :param traceConfig: 
        :type traceConfig: TraceConfig
        """
        return (
            cls.build_send_payload("start", {
                "categories": categories,
                "options": options,
                "bufferUsageReportingInterval": bufferUsageReportingInterval,
                "transferMode": transferMode,
                "streamCompression": streamCompression,
                "traceConfig": traceConfig,
            }),
            None
        )



class BufferUsageEvent(BaseEvent):

    js_name = 'Tracing.bufferUsage'
    hashable = []
    is_hashable = False

    def __init__(self,
                 percentFull: Union['float', dict, None] = None,
                 eventCount: Union['float', dict, None] = None,
                 value: Union['float', dict, None] = None,
                 ):
        if isinstance(percentFull, dict):
            percentFull = float(**percentFull)
        self.percentFull = percentFull
        if isinstance(eventCount, dict):
            eventCount = float(**eventCount)
        self.eventCount = eventCount
        if isinstance(value, dict):
            value = float(**value)
        self.value = value

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class DataCollectedEvent(BaseEvent):

    js_name = 'Tracing.dataCollected'
    hashable = []
    is_hashable = False

    def __init__(self,
                 value: Union['[]', dict],
                 ):
        if isinstance(value, dict):
            value = [](**value)
        self.value = value

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class TracingCompleteEvent(BaseEvent):

    js_name = 'Tracing.tracingComplete'
    hashable = []
    is_hashable = False

    def __init__(self,
                 stream: Union['IO.StreamHandle', dict, None] = None,
                 streamCompression: Union['StreamCompression', dict, None] = None,
                 ):
        if isinstance(stream, dict):
            stream = IO.StreamHandle(**stream)
        self.stream = stream
        if isinstance(streamCompression, dict):
            streamCompression = StreamCompression(**streamCompression)
        self.streamCompression = streamCompression

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import dom as DOM
from chromewhip.protocol import debugger as Debugger

# DOMBreakpointType: DOM breakpoint type.
DOMBreakpointType = str

# EventListener: Object event listener.
class EventListener(ChromeTypeBase):
    def __init__(self,
                 type: Union['str'],
                 useCapture: Union['bool'],
                 passive: Union['bool'],
                 once: Union['bool'],
                 scriptId: Union['Runtime.ScriptId'],
                 lineNumber: Union['int'],
                 columnNumber: Union['int'],
                 handler: Optional['Runtime.RemoteObject'] = None,
                 originalHandler: Optional['Runtime.RemoteObject'] = None,
                 backendNodeId: Optional['DOM.BackendNodeId'] = None,
                 ):

        self.type = type
        self.useCapture = useCapture
        self.passive = passive
        self.once = once
        self.scriptId = scriptId
        self.lineNumber = lineNumber
        self.columnNumber = columnNumber
        self.handler = handler
        self.originalHandler = originalHandler
        self.backendNodeId = backendNodeId


class DOMDebugger(PayloadMixin):
    """ DOM debugging allows setting breakpoints on particular DOM operations and events. JavaScript
execution will stop on these operations as if there was a regular breakpoint set.
    """
    @classmethod
    def getEventListeners(cls,
                          objectId: Union['Runtime.RemoteObjectId'],
                          depth: Optional['int'] = None,
                          pierce: Optional['bool'] = None,
                          ):
        """Returns event listeners of the given object.
        :param objectId: Identifier of the object to return listeners for.
        :type objectId: Runtime.RemoteObjectId
        :param depth: The maximum depth at which Node children should be retrieved, defaults to 1. Use -1 for the
entire subtree or provide an integer larger than 0.
        :type depth: int
        :param pierce: Whether or not iframes and shadow roots should be traversed when returning the subtree
(default is false). Reports listeners for all contexts if pierce is enabled.
        :type pierce: bool
        """
        return (
            cls.build_send_payload("getEventListeners", {
                "objectId": objectId,
                "depth": depth,
                "pierce": pierce,
            }),
            cls.convert_payload({
                "listeners": {
                    "class": [EventListener],
                    "optional": False
                },
            })
        )

    @classmethod
    def removeDOMBreakpoint(cls,
                            nodeId: Union['DOM.NodeId'],
                            type: Union['DOMBreakpointType'],
                            ):
        """Removes DOM breakpoint that was set using `setDOMBreakpoint`.
        :param nodeId: Identifier of the node to remove breakpoint from.
        :type nodeId: DOM.NodeId
        :param type: Type of the breakpoint to remove.
        :type type: DOMBreakpointType
        """
        return (
            cls.build_send_payload("removeDOMBreakpoint", {
                "nodeId": nodeId,
                "type": type,
            }),
            None
        )

    @classmethod
    def removeEventListenerBreakpoint(cls,
                                      eventName: Union['str'],
                                      targetName: Optional['str'] = None,
                                      ):
        """Removes breakpoint on particular DOM event.
        :param eventName: Event name.
        :type eventName: str
        :param targetName: EventTarget interface name.
        :type targetName: str
        """
        return (
            cls.build_send_payload("removeEventListenerBreakpoint", {
                "eventName": eventName,
                "targetName": targetName,
            }),
            None
        )

    @classmethod
    def removeInstrumentationBreakpoint(cls,
                                        eventName: Union['str'],
                                        ):
        """Removes breakpoint on particular native event.
        :param eventName: Instrumentation name to stop on.
        :type eventName: str
        """
        return (
            cls.build_send_payload("removeInstrumentationBreakpoint", {
                "eventName": eventName,
            }),
            None
        )

    @classmethod
    def removeXHRBreakpoint(cls,
                            url: Union['str'],
                            ):
        """Removes breakpoint from XMLHttpRequest.
        :param url: Resource URL substring.
        :type url: str
        """
        return (
            cls.build_send_payload("removeXHRBreakpoint", {
                "url": url,
            }),
            None
        )

    @classmethod
    def setDOMBreakpoint(cls,
                         nodeId: Union['DOM.NodeId'],
                         type: Union['DOMBreakpointType'],
                         ):
        """Sets breakpoint on particular operation with DOM.
        :param nodeId: Identifier of the node to set breakpoint on.
        :type nodeId: DOM.NodeId
        :param type: Type of the operation to stop upon.
        :type type: DOMBreakpointType
        """
        return (
            cls.build_send_payload("setDOMBreakpoint", {
                "nodeId": nodeId,
                "type": type,
            }),
            None
        )

    @classmethod
    def setEventListenerBreakpoint(cls,
                                   eventName: Union['str'],
                                   targetName: Optional['str'] = None,
                                   ):
        """Sets breakpoint on particular DOM event.
        :param eventName: DOM Event name to stop on (any DOM event will do).
        :type eventName: str
        :param targetName: EventTarget interface name to stop on. If equal to `"*"` or not provided, will stop on any
EventTarget.
        :type targetName: str
        """
        return (
            cls.build_send_payload("setEventListenerBreakpoint", {
                "eventName": eventName,
                "targetName": targetName,
            }),
            None
        )

    @classmethod
    def setInstrumentationBreakpoint(cls,
                                     eventName: Union['str'],
                                     ):
        """Sets breakpoint on particular native event.
        :param eventName: Instrumentation name to stop on.
        :type eventName: str
        """
        return (
            cls.build_send_payload("setInstrumentationBreakpoint", {
                "eventName": eventName,
            }),
            None
        )

    @classmethod
    def setXHRBreakpoint(cls,
                         url: Union['str'],
                         ):
        """Sets breakpoint on XMLHttpRequest.
        :param url: Resource URL substring. All XHRs having this substring in the URL will get stopped upon.
        :type url: str
        """
        return (
            cls.build_send_payload("setXHRBreakpoint", {
                "url": url,
            }),
            None
        )




# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import runtime as Runtime

# HeapSnapshotObjectId: Heap snapshot object id.
HeapSnapshotObjectId = str

# SamplingHeapProfileNode: Sampling Heap Profile node. Holds callsite information, allocation statistics and child nodes.
class SamplingHeapProfileNode(ChromeTypeBase):
    def __init__(self,
                 callFrame: Union['Runtime.CallFrame'],
                 selfSize: Union['float'],
                 children: Union['[SamplingHeapProfileNode]'],
                 ):

        self.callFrame = callFrame
        self.selfSize = selfSize
        self.children = children


# SamplingHeapProfile: Profile.
class SamplingHeapProfile(ChromeTypeBase):
    def __init__(self,
                 head: Union['SamplingHeapProfileNode'],
                 ):

        self.head = head


class HeapProfiler(PayloadMixin):
    """ 
    """
    @classmethod
    def addInspectedHeapObject(cls,
                               heapObjectId: Union['HeapSnapshotObjectId'],
                               ):
        """Enables console to refer to the node with given id via $x (see Command Line API for more details
$x functions).
        :param heapObjectId: Heap snapshot object id to be accessible by means of $x command line API.
        :type heapObjectId: HeapSnapshotObjectId
        """
        return (
            cls.build_send_payload("addInspectedHeapObject", {
                "heapObjectId": heapObjectId,
            }),
            None
        )

    @classmethod
    def collectGarbage(cls):
        """
        """
        return (
            cls.build_send_payload("collectGarbage", {
            }),
            None
        )

    @classmethod
    def disable(cls):
        """
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def getHeapObjectId(cls,
                        objectId: Union['Runtime.RemoteObjectId'],
                        ):
        """
        :param objectId: Identifier of the object to get heap object id for.
        :type objectId: Runtime.RemoteObjectId
        """
        return (
            cls.build_send_payload("getHeapObjectId", {
                "objectId": objectId,
            }),
            cls.convert_payload({
                "heapSnapshotObjectId": {
                    "class": HeapSnapshotObjectId,
                    "optional": False
                },
            })
        )

    @classmethod
    def getObjectByHeapObjectId(cls,
                                objectId: Union['HeapSnapshotObjectId'],
                                objectGroup: Optional['str'] = None,
                                ):
        """
        :param objectId: 
        :type objectId: HeapSnapshotObjectId
        :param objectGroup: Symbolic group name that can be used to release multiple objects.
        :type objectGroup: str
        """
        return (
            cls.build_send_payload("getObjectByHeapObjectId", {
                "objectId": objectId,
                "objectGroup": objectGroup,
            }),
            cls.convert_payload({
                "result": {
                    "class": Runtime.RemoteObject,
                    "optional": False
                },
            })
        )

    @classmethod
    def getSamplingProfile(cls):
        """
        """
        return (
            cls.build_send_payload("getSamplingProfile", {
            }),
            cls.convert_payload({
                "profile": {
                    "class": SamplingHeapProfile,
                    "optional": False
                },
            })
        )

    @classmethod
    def startSampling(cls,
                      samplingInterval: Optional['float'] = None,
                      ):
        """
        :param samplingInterval: Average sample interval in bytes. Poisson distribution is used for the intervals. The
default value is 32768 bytes.
        :type samplingInterval: float
        """
        return (
            cls.build_send_payload("startSampling", {
                "samplingInterval": samplingInterval,
            }),
            None
        )

    @classmethod
    def startTrackingHeapObjects(cls,
                                 trackAllocations: Optional['bool'] = None,
                                 ):
        """
        :param trackAllocations: 
        :type trackAllocations: bool
        """
        return (
            cls.build_send_payload("startTrackingHeapObjects", {
                "trackAllocations": trackAllocations,
            }),
            None
        )

    @classmethod
    def stopSampling(cls):
        """
        """
        return (
            cls.build_send_payload("stopSampling", {
            }),
            cls.convert_payload({
                "profile": {
                    "class": SamplingHeapProfile,
                    "optional": False
                },
            })
        )

    @classmethod
    def stopTrackingHeapObjects(cls,
                                reportProgress: Optional['bool'] = None,
                                ):
        """
        :param reportProgress: If true 'reportHeapSnapshotProgress' events will be generated while snapshot is being taken
when the tracking is stopped.
        :type reportProgress: bool
        """
        return (
            cls.build_send_payload("stopTrackingHeapObjects", {
                "reportProgress": reportProgress,
            }),
            None
        )

    @classmethod
    def takeHeapSnapshot(cls,
                         reportProgress: Optional['bool'] = None,
                         ):
        """
        :param reportProgress: If true 'reportHeapSnapshotProgress' events will be generated while snapshot is being taken.
        :type reportProgress: bool
        """
        return (
            cls.build_send_payload("takeHeapSnapshot", {
                "reportProgress": reportProgress,
            }),
            None
        )



class AddHeapSnapshotChunkEvent(BaseEvent):

    js_name = 'Heapprofiler.addHeapSnapshotChunk'
    hashable = []
    is_hashable = False

    def __init__(self,
                 chunk: Union['str', dict],
                 ):
        if isinstance(chunk, dict):
            chunk = str(**chunk)
        self.chunk = chunk

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class HeapStatsUpdateEvent(BaseEvent):

    js_name = 'Heapprofiler.heapStatsUpdate'
    hashable = []
    is_hashable = False

    def __init__(self,
                 statsUpdate: Union['[]', dict],
                 ):
        if isinstance(statsUpdate, dict):
            statsUpdate = [](**statsUpdate)
        self.statsUpdate = statsUpdate

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class LastSeenObjectIdEvent(BaseEvent):

    js_name = 'Heapprofiler.lastSeenObjectId'
    hashable = ['lastSeenObjectId']
    is_hashable = True

    def __init__(self,
                 lastSeenObjectId: Union['int', dict],
                 timestamp: Union['float', dict],
                 ):
        if isinstance(lastSeenObjectId, dict):
            lastSeenObjectId = int(**lastSeenObjectId)
        self.lastSeenObjectId = lastSeenObjectId
        if isinstance(timestamp, dict):
            timestamp = float(**timestamp)
        self.timestamp = timestamp

    @classmethod
    def build_hash(cls, lastSeenObjectId):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ReportHeapSnapshotProgressEvent(BaseEvent):

    js_name = 'Heapprofiler.reportHeapSnapshotProgress'
    hashable = []
    is_hashable = False

    def __init__(self,
                 done: Union['int', dict],
                 total: Union['int', dict],
                 finished: Union['bool', dict, None] = None,
                 ):
        if isinstance(done, dict):
            done = int(**done)
        self.done = done
        if isinstance(total, dict):
            total = int(**total)
        self.total = total
        if isinstance(finished, dict):
            finished = bool(**finished)
        self.finished = finished

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class ResetProfilesEvent(BaseEvent):

    js_name = 'Heapprofiler.resetProfiles'
    hashable = []
    is_hashable = False

    def __init__(self):
        pass

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



import asyncio
import functools
import logging

from bs4 import BeautifulSoup
from aiohttp import web

from chromewhip.protocol import page, emulation, browser, dom, runtime

BS = functools.partial(BeautifulSoup, features="lxml")

log = logging.getLogger('chromewhip.views')

async def _go(request: web.Request):

    js_profiles = request.app['js-profiles']
    c = request.app['chrome-driver']

    url = request.query.get('url')
    if not url:
        return web.HTTPBadRequest(reason='no url query param provided')  # TODO: match splash reply

    wait_s = float(request.query.get('wait', 0))

    raw_viewport = request.query.get('viewport', '1024x768')
    parts = raw_viewport.split('x')
    width = int(parts[0])
    height = int(parts[1])

    js_profile_name = request.query.get('js', None)
    if js_profile_name:
        profile = js_profiles.get(js_profile_name)
        if not profile:
            return web.HTTPBadRequest(reason='profile name is incorrect')  # TODO: match splash

    # TODO: potentially validate and verify js source for errors and security concerrns
    js_source = request.query.get('js_source', None)

    await c.connect()
    tab = c.tabs[0]
    cmd = page.Page.setDeviceMetricsOverride(width=width,
                                             height=height,
                                             deviceScaleFactor=0.0,
                                             mobile=False)
    await tab.send_command(cmd)
    await tab.enable_page_events()
    await tab.go(url)
    await asyncio.sleep(wait_s)
    if js_profile_name:
        await tab.evaluate(js_profiles[js_profile_name])

    if js_source:
        await tab.evaluate(js_source)

    return tab


async def render_html(request: web.Request):
    # https://splash.readthedocs.io/en/stable/api.html#render-html
    tab = await _go(request)
    return web.Response(text=BS((await tab.html()).decode()).prettify())


async def render_png(request: web.Request):
    # https://splash.readthedocs.io/en/stable/api.html#render-png
    tab = await _go(request)

    should_render_all = True if request.query.get('render_all', False) == '1' else False

    if not should_render_all:
        data = await tab.screenshot()
        return web.Response(body=data, content_type='image/png')

    if should_render_all:
        raw_viewport = request.query.get('viewport', '1024x768')
        parts = raw_viewport.split('x')
        width = int(parts[0])
        height = int(parts[1])
        cmd = page.Page.setDeviceMetricsOverride(width=int(width),
                                                 height=int(height),
                                                 deviceScaleFactor=0.0,
                                                 mobile=False)
        await tab.send_command(cmd)

        # model numbers affected by device metrics, so needs to come after
        res = await tab.send_command(dom.DOM.getDocument())
        doc_node_id = res['ack']['result']['root'].nodeId
        res = await tab.send_command(dom.DOM.querySelector(selector='body', nodeId=doc_node_id))
        body_node_id = res['ack']['result']['nodeId']
        res = await tab.send_command(dom.DOM.getBoxModel(nodeId=body_node_id))
        full_height = res['ack']['result']['model'].height
        log.debug('full_height = %s' % full_height)

        offset = 0
        import base64
        from PIL import Image
        from io import BytesIO
        full_image = Image.new('RGB', (int(width), int(full_height)))
        delta = int(height)
        while offset < full_height + 1:  # TODO: cut+paste to exact dimensions
            await tab.send_command(runtime.Runtime.evaluate('window.scrollTo(0, %s)' % offset))
            result = await tab.send_command(page.Page.captureScreenshot(format='png', fromSurface=False))
            base64_data = result['ack']['result']['data']
            snapshot = Image.open(BytesIO(base64.b64decode(base64_data)))
            full_image.paste(snapshot, (0, offset))
            offset += delta
        output = BytesIO()
        full_image.save(output, format='png')
        return web.Response(body=output.getvalue(), content_type='image/png')




# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# GPUDevice: Describes a single graphics processor (GPU).
class GPUDevice(ChromeTypeBase):
    def __init__(self,
                 vendorId: Union['float'],
                 deviceId: Union['float'],
                 vendorString: Union['str'],
                 deviceString: Union['str'],
                 ):

        self.vendorId = vendorId
        self.deviceId = deviceId
        self.vendorString = vendorString
        self.deviceString = deviceString


# GPUInfo: Provides information about the GPU(s) on the system.
class GPUInfo(ChromeTypeBase):
    def __init__(self,
                 devices: Union['[GPUDevice]'],
                 driverBugWorkarounds: Union['[]'],
                 auxAttributes: Optional['dict'] = None,
                 featureStatus: Optional['dict'] = None,
                 ):

        self.devices = devices
        self.auxAttributes = auxAttributes
        self.featureStatus = featureStatus
        self.driverBugWorkarounds = driverBugWorkarounds


class SystemInfo(PayloadMixin):
    """ The SystemInfo domain defines methods and events for querying low-level system information.
    """
    @classmethod
    def getInfo(cls):
        """Returns information about the system.
        """
        return (
            cls.build_send_payload("getInfo", {
            }),
            cls.convert_payload({
                "gpu": {
                    "class": GPUInfo,
                    "optional": False
                },
                "modelName": {
                    "class": str,
                    "optional": False
                },
                "modelVersion": {
                    "class": str,
                    "optional": False
                },
                "commandLine": {
                    "class": str,
                    "optional": False
                },
            })
        )




# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import css as CSS
from chromewhip.protocol import dom as DOM
from chromewhip.protocol import domdebugger as DOMDebugger
from chromewhip.protocol import page as Page

# DOMNode: A Node in the DOM tree.
class DOMNode(ChromeTypeBase):
    def __init__(self,
                 nodeType: Union['int'],
                 nodeName: Union['str'],
                 nodeValue: Union['str'],
                 backendNodeId: Union['DOM.BackendNodeId'],
                 textValue: Optional['str'] = None,
                 inputValue: Optional['str'] = None,
                 inputChecked: Optional['bool'] = None,
                 optionSelected: Optional['bool'] = None,
                 childNodeIndexes: Optional['[]'] = None,
                 attributes: Optional['[NameValue]'] = None,
                 pseudoElementIndexes: Optional['[]'] = None,
                 layoutNodeIndex: Optional['int'] = None,
                 documentURL: Optional['str'] = None,
                 baseURL: Optional['str'] = None,
                 contentLanguage: Optional['str'] = None,
                 documentEncoding: Optional['str'] = None,
                 publicId: Optional['str'] = None,
                 systemId: Optional['str'] = None,
                 frameId: Optional['Page.FrameId'] = None,
                 contentDocumentIndex: Optional['int'] = None,
                 pseudoType: Optional['DOM.PseudoType'] = None,
                 shadowRootType: Optional['DOM.ShadowRootType'] = None,
                 isClickable: Optional['bool'] = None,
                 eventListeners: Optional['[DOMDebugger.EventListener]'] = None,
                 currentSourceURL: Optional['str'] = None,
                 originURL: Optional['str'] = None,
                 ):

        self.nodeType = nodeType
        self.nodeName = nodeName
        self.nodeValue = nodeValue
        self.textValue = textValue
        self.inputValue = inputValue
        self.inputChecked = inputChecked
        self.optionSelected = optionSelected
        self.backendNodeId = backendNodeId
        self.childNodeIndexes = childNodeIndexes
        self.attributes = attributes
        self.pseudoElementIndexes = pseudoElementIndexes
        self.layoutNodeIndex = layoutNodeIndex
        self.documentURL = documentURL
        self.baseURL = baseURL
        self.contentLanguage = contentLanguage
        self.documentEncoding = documentEncoding
        self.publicId = publicId
        self.systemId = systemId
        self.frameId = frameId
        self.contentDocumentIndex = contentDocumentIndex
        self.pseudoType = pseudoType
        self.shadowRootType = shadowRootType
        self.isClickable = isClickable
        self.eventListeners = eventListeners
        self.currentSourceURL = currentSourceURL
        self.originURL = originURL


# InlineTextBox: Details of post layout rendered text positions. The exact layout should not be regarded asstable and may change between versions.
class InlineTextBox(ChromeTypeBase):
    def __init__(self,
                 boundingBox: Union['DOM.Rect'],
                 startCharacterIndex: Union['int'],
                 numCharacters: Union['int'],
                 ):

        self.boundingBox = boundingBox
        self.startCharacterIndex = startCharacterIndex
        self.numCharacters = numCharacters


# LayoutTreeNode: Details of an element in the DOM tree with a LayoutObject.
class LayoutTreeNode(ChromeTypeBase):
    def __init__(self,
                 domNodeIndex: Union['int'],
                 boundingBox: Union['DOM.Rect'],
                 layoutText: Optional['str'] = None,
                 inlineTextNodes: Optional['[InlineTextBox]'] = None,
                 styleIndex: Optional['int'] = None,
                 paintOrder: Optional['int'] = None,
                 ):

        self.domNodeIndex = domNodeIndex
        self.boundingBox = boundingBox
        self.layoutText = layoutText
        self.inlineTextNodes = inlineTextNodes
        self.styleIndex = styleIndex
        self.paintOrder = paintOrder


# ComputedStyle: A subset of the full ComputedStyle as defined by the request whitelist.
class ComputedStyle(ChromeTypeBase):
    def __init__(self,
                 properties: Union['[NameValue]'],
                 ):

        self.properties = properties


# NameValue: A name/value pair.
class NameValue(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 value: Union['str'],
                 ):

        self.name = name
        self.value = value


# StringIndex: Index of the string in the strings table.
StringIndex = int

# ArrayOfStrings: Index of the string in the strings table.
ArrayOfStrings = [StringIndex]

# RareStringData: Data that is only present on rare nodes.
class RareStringData(ChromeTypeBase):
    def __init__(self,
                 index: Union['[]'],
                 value: Union['[StringIndex]'],
                 ):

        self.index = index
        self.value = value


# RareBooleanData: 
class RareBooleanData(ChromeTypeBase):
    def __init__(self,
                 index: Union['[]'],
                 ):

        self.index = index


# RareIntegerData: 
class RareIntegerData(ChromeTypeBase):
    def __init__(self,
                 index: Union['[]'],
                 value: Union['[]'],
                 ):

        self.index = index
        self.value = value


# Rectangle: 
Rectangle = [float]

# DocumentSnapshot: Document snapshot.
class DocumentSnapshot(ChromeTypeBase):
    def __init__(self,
                 documentURL: Union['StringIndex'],
                 baseURL: Union['StringIndex'],
                 contentLanguage: Union['StringIndex'],
                 encodingName: Union['StringIndex'],
                 publicId: Union['StringIndex'],
                 systemId: Union['StringIndex'],
                 frameId: Union['StringIndex'],
                 nodes: Union['NodeTreeSnapshot'],
                 layout: Union['LayoutTreeSnapshot'],
                 textBoxes: Union['TextBoxSnapshot'],
                 ):

        self.documentURL = documentURL
        self.baseURL = baseURL
        self.contentLanguage = contentLanguage
        self.encodingName = encodingName
        self.publicId = publicId
        self.systemId = systemId
        self.frameId = frameId
        self.nodes = nodes
        self.layout = layout
        self.textBoxes = textBoxes


# NodeTreeSnapshot: Table containing nodes.
class NodeTreeSnapshot(ChromeTypeBase):
    def __init__(self,
                 parentIndex: Optional['[]'] = None,
                 nodeType: Optional['[]'] = None,
                 nodeName: Optional['[StringIndex]'] = None,
                 nodeValue: Optional['[StringIndex]'] = None,
                 backendNodeId: Optional['[DOM.BackendNodeId]'] = None,
                 attributes: Optional['[ArrayOfStrings]'] = None,
                 textValue: Optional['RareStringData'] = None,
                 inputValue: Optional['RareStringData'] = None,
                 inputChecked: Optional['RareBooleanData'] = None,
                 optionSelected: Optional['RareBooleanData'] = None,
                 contentDocumentIndex: Optional['RareIntegerData'] = None,
                 pseudoType: Optional['RareStringData'] = None,
                 isClickable: Optional['RareBooleanData'] = None,
                 currentSourceURL: Optional['RareStringData'] = None,
                 originURL: Optional['RareStringData'] = None,
                 ):

        self.parentIndex = parentIndex
        self.nodeType = nodeType
        self.nodeName = nodeName
        self.nodeValue = nodeValue
        self.backendNodeId = backendNodeId
        self.attributes = attributes
        self.textValue = textValue
        self.inputValue = inputValue
        self.inputChecked = inputChecked
        self.optionSelected = optionSelected
        self.contentDocumentIndex = contentDocumentIndex
        self.pseudoType = pseudoType
        self.isClickable = isClickable
        self.currentSourceURL = currentSourceURL
        self.originURL = originURL


# LayoutTreeSnapshot: Details of an element in the DOM tree with a LayoutObject.
class LayoutTreeSnapshot(ChromeTypeBase):
    def __init__(self,
                 nodeIndex: Union['[]'],
                 styles: Union['[ArrayOfStrings]'],
                 bounds: Union['[Rectangle]'],
                 text: Union['[StringIndex]'],
                 ):

        self.nodeIndex = nodeIndex
        self.styles = styles
        self.bounds = bounds
        self.text = text


# TextBoxSnapshot: Details of post layout rendered text positions. The exact layout should not be regarded asstable and may change between versions.
class TextBoxSnapshot(ChromeTypeBase):
    def __init__(self,
                 layoutIndex: Union['[]'],
                 bounds: Union['[Rectangle]'],
                 start: Union['[]'],
                 length: Union['[]'],
                 ):

        self.layoutIndex = layoutIndex
        self.bounds = bounds
        self.start = start
        self.length = length


class DOMSnapshot(PayloadMixin):
    """ This domain facilitates obtaining document snapshots with DOM, layout, and style information.
    """
    @classmethod
    def disable(cls):
        """Disables DOM snapshot agent for the given page.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables DOM snapshot agent for the given page.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def getSnapshot(cls,
                    computedStyleWhitelist: Union['[]'],
                    includeEventListeners: Optional['bool'] = None,
                    includePaintOrder: Optional['bool'] = None,
                    includeUserAgentShadowTree: Optional['bool'] = None,
                    ):
        """Returns a document snapshot, including the full DOM tree of the root node (including iframes,
template contents, and imported documents) in a flattened array, as well as layout and
white-listed computed style information for the nodes. Shadow DOM in the returned DOM tree is
flattened.
        :param computedStyleWhitelist: Whitelist of computed styles to return.
        :type computedStyleWhitelist: []
        :param includeEventListeners: Whether or not to retrieve details of DOM listeners (default false).
        :type includeEventListeners: bool
        :param includePaintOrder: Whether to determine and include the paint order index of LayoutTreeNodes (default false).
        :type includePaintOrder: bool
        :param includeUserAgentShadowTree: Whether to include UA shadow tree in the snapshot (default false).
        :type includeUserAgentShadowTree: bool
        """
        return (
            cls.build_send_payload("getSnapshot", {
                "computedStyleWhitelist": computedStyleWhitelist,
                "includeEventListeners": includeEventListeners,
                "includePaintOrder": includePaintOrder,
                "includeUserAgentShadowTree": includeUserAgentShadowTree,
            }),
            cls.convert_payload({
                "domNodes": {
                    "class": [DOMNode],
                    "optional": False
                },
                "layoutTreeNodes": {
                    "class": [LayoutTreeNode],
                    "optional": False
                },
                "computedStyles": {
                    "class": [ComputedStyle],
                    "optional": False
                },
            })
        )

    @classmethod
    def captureSnapshot(cls,
                        computedStyles: Union['[]'],
                        ):
        """Returns a document snapshot, including the full DOM tree of the root node (including iframes,
template contents, and imported documents) in a flattened array, as well as layout and
white-listed computed style information for the nodes. Shadow DOM in the returned DOM tree is
flattened.
        :param computedStyles: Whitelist of computed styles to return.
        :type computedStyles: []
        """
        return (
            cls.build_send_payload("captureSnapshot", {
                "computedStyles": computedStyles,
            }),
            cls.convert_payload({
                "documents": {
                    "class": [DocumentSnapshot],
                    "optional": False
                },
                "strings": {
                    "class": [],
                    "optional": False
                },
            })
        )




# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import runtime as Runtime
from chromewhip.protocol import debugger as Debugger

# ProfileNode: Profile node. Holds callsite information, execution statistics and child nodes.
class ProfileNode(ChromeTypeBase):
    def __init__(self,
                 id: Union['int'],
                 callFrame: Union['Runtime.CallFrame'],
                 hitCount: Optional['int'] = None,
                 children: Optional['[]'] = None,
                 deoptReason: Optional['str'] = None,
                 positionTicks: Optional['[PositionTickInfo]'] = None,
                 ):

        self.id = id
        self.callFrame = callFrame
        self.hitCount = hitCount
        self.children = children
        self.deoptReason = deoptReason
        self.positionTicks = positionTicks


# Profile: Profile.
class Profile(ChromeTypeBase):
    def __init__(self,
                 nodes: Union['[ProfileNode]'],
                 startTime: Union['float'],
                 endTime: Union['float'],
                 samples: Optional['[]'] = None,
                 timeDeltas: Optional['[]'] = None,
                 ):

        self.nodes = nodes
        self.startTime = startTime
        self.endTime = endTime
        self.samples = samples
        self.timeDeltas = timeDeltas


# PositionTickInfo: Specifies a number of samples attributed to a certain source position.
class PositionTickInfo(ChromeTypeBase):
    def __init__(self,
                 line: Union['int'],
                 ticks: Union['int'],
                 ):

        self.line = line
        self.ticks = ticks


# CoverageRange: Coverage data for a source range.
class CoverageRange(ChromeTypeBase):
    def __init__(self,
                 startOffset: Union['int'],
                 endOffset: Union['int'],
                 count: Union['int'],
                 ):

        self.startOffset = startOffset
        self.endOffset = endOffset
        self.count = count


# FunctionCoverage: Coverage data for a JavaScript function.
class FunctionCoverage(ChromeTypeBase):
    def __init__(self,
                 functionName: Union['str'],
                 ranges: Union['[CoverageRange]'],
                 isBlockCoverage: Union['bool'],
                 ):

        self.functionName = functionName
        self.ranges = ranges
        self.isBlockCoverage = isBlockCoverage


# ScriptCoverage: Coverage data for a JavaScript script.
class ScriptCoverage(ChromeTypeBase):
    def __init__(self,
                 scriptId: Union['Runtime.ScriptId'],
                 url: Union['str'],
                 functions: Union['[FunctionCoverage]'],
                 ):

        self.scriptId = scriptId
        self.url = url
        self.functions = functions


# TypeObject: Describes a type collected during runtime.
class TypeObject(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 ):

        self.name = name


# TypeProfileEntry: Source offset and types for a parameter or return value.
class TypeProfileEntry(ChromeTypeBase):
    def __init__(self,
                 offset: Union['int'],
                 types: Union['[TypeObject]'],
                 ):

        self.offset = offset
        self.types = types


# ScriptTypeProfile: Type profile data collected during runtime for a JavaScript script.
class ScriptTypeProfile(ChromeTypeBase):
    def __init__(self,
                 scriptId: Union['Runtime.ScriptId'],
                 url: Union['str'],
                 entries: Union['[TypeProfileEntry]'],
                 ):

        self.scriptId = scriptId
        self.url = url
        self.entries = entries


class Profiler(PayloadMixin):
    """ 
    """
    @classmethod
    def disable(cls):
        """
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def getBestEffortCoverage(cls):
        """Collect coverage data for the current isolate. The coverage data may be incomplete due to
garbage collection.
        """
        return (
            cls.build_send_payload("getBestEffortCoverage", {
            }),
            cls.convert_payload({
                "result": {
                    "class": [ScriptCoverage],
                    "optional": False
                },
            })
        )

    @classmethod
    def setSamplingInterval(cls,
                            interval: Union['int'],
                            ):
        """Changes CPU profiler sampling interval. Must be called before CPU profiles recording started.
        :param interval: New sampling interval in microseconds.
        :type interval: int
        """
        return (
            cls.build_send_payload("setSamplingInterval", {
                "interval": interval,
            }),
            None
        )

    @classmethod
    def start(cls):
        """
        """
        return (
            cls.build_send_payload("start", {
            }),
            None
        )

    @classmethod
    def startPreciseCoverage(cls,
                             callCount: Optional['bool'] = None,
                             detailed: Optional['bool'] = None,
                             ):
        """Enable precise code coverage. Coverage data for JavaScript executed before enabling precise code
coverage may be incomplete. Enabling prevents running optimized code and resets execution
counters.
        :param callCount: Collect accurate call counts beyond simple 'covered' or 'not covered'.
        :type callCount: bool
        :param detailed: Collect block-based coverage.
        :type detailed: bool
        """
        return (
            cls.build_send_payload("startPreciseCoverage", {
                "callCount": callCount,
                "detailed": detailed,
            }),
            None
        )

    @classmethod
    def startTypeProfile(cls):
        """Enable type profile.
        """
        return (
            cls.build_send_payload("startTypeProfile", {
            }),
            None
        )

    @classmethod
    def stop(cls):
        """
        """
        return (
            cls.build_send_payload("stop", {
            }),
            cls.convert_payload({
                "profile": {
                    "class": Profile,
                    "optional": False
                },
            })
        )

    @classmethod
    def stopPreciseCoverage(cls):
        """Disable precise code coverage. Disabling releases unnecessary execution count records and allows
executing optimized code.
        """
        return (
            cls.build_send_payload("stopPreciseCoverage", {
            }),
            None
        )

    @classmethod
    def stopTypeProfile(cls):
        """Disable type profile. Disabling releases type profile data collected so far.
        """
        return (
            cls.build_send_payload("stopTypeProfile", {
            }),
            None
        )

    @classmethod
    def takePreciseCoverage(cls):
        """Collect coverage data for the current isolate, and resets execution counters. Precise code
coverage needs to have started.
        """
        return (
            cls.build_send_payload("takePreciseCoverage", {
            }),
            cls.convert_payload({
                "result": {
                    "class": [ScriptCoverage],
                    "optional": False
                },
            })
        )

    @classmethod
    def takeTypeProfile(cls):
        """Collect type profile.
        """
        return (
            cls.build_send_payload("takeTypeProfile", {
            }),
            cls.convert_payload({
                "result": {
                    "class": [ScriptTypeProfile],
                    "optional": False
                },
            })
        )



class ConsoleProfileFinishedEvent(BaseEvent):

    js_name = 'Profiler.consoleProfileFinished'
    hashable = ['id']
    is_hashable = True

    def __init__(self,
                 id: Union['str', dict],
                 location: Union['Debugger.Location', dict],
                 profile: Union['Profile', dict],
                 title: Union['str', dict, None] = None,
                 ):
        if isinstance(id, dict):
            id = str(**id)
        self.id = id
        if isinstance(location, dict):
            location = Debugger.Location(**location)
        self.location = location
        if isinstance(profile, dict):
            profile = Profile(**profile)
        self.profile = profile
        if isinstance(title, dict):
            title = str(**title)
        self.title = title

    @classmethod
    def build_hash(cls, id):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h


class ConsoleProfileStartedEvent(BaseEvent):

    js_name = 'Profiler.consoleProfileStarted'
    hashable = ['id']
    is_hashable = True

    def __init__(self,
                 id: Union['str', dict],
                 location: Union['Debugger.Location', dict],
                 title: Union['str', dict, None] = None,
                 ):
        if isinstance(id, dict):
            id = str(**id)
        self.id = id
        if isinstance(location, dict):
            location = Debugger.Location(**location)
        self.location = location
        if isinstance(title, dict):
            title = str(**title)
        self.title = title

    @classmethod
    def build_hash(cls, id):
        kwargs = locals()
        kwargs.pop('cls')
        serialized_id_params = ','.join(['='.join([p, str(v)]) for p, v in kwargs.items()])
        h = '{}:{}'.format(cls.js_name, serialized_id_params)
        log.debug('generated hash = %s' % h)
        return h



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import runtime as Runtime

# ConsoleMessage: Console message.
class ConsoleMessage(ChromeTypeBase):
    def __init__(self,
                 source: Union['str'],
                 level: Union['str'],
                 text: Union['str'],
                 url: Optional['str'] = None,
                 line: Optional['int'] = None,
                 column: Optional['int'] = None,
                 ):

        self.source = source
        self.level = level
        self.text = text
        self.url = url
        self.line = line
        self.column = column


class Console(PayloadMixin):
    """ This domain is deprecated - use Runtime or Log instead.
    """
    @classmethod
    def clearMessages(cls):
        """Does nothing.
        """
        return (
            cls.build_send_payload("clearMessages", {
            }),
            None
        )

    @classmethod
    def disable(cls):
        """Disables console domain, prevents further console messages from being reported to the client.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables console domain, sends the messages collected so far to the client by means of the
`messageAdded` notification.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )



class MessageAddedEvent(BaseEvent):

    js_name = 'Console.messageAdded'
    hashable = []
    is_hashable = False

    def __init__(self,
                 message: Union['ConsoleMessage', dict],
                 ):
        if isinstance(message, dict):
            message = ConsoleMessage(**message)
        self.message = message

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# StreamHandle: This is either obtained from another method or specifed as `blob:&lt;uuid&gt;` where`&lt;uuid&gt` is an UUID of a Blob.
StreamHandle = str

class IO(PayloadMixin):
    """ Input/Output operations for streams produced by DevTools.
    """
    @classmethod
    def close(cls,
              handle: Union['StreamHandle'],
              ):
        """Close the stream, discard any temporary backing storage.
        :param handle: Handle of the stream to close.
        :type handle: StreamHandle
        """
        return (
            cls.build_send_payload("close", {
                "handle": handle,
            }),
            None
        )

    @classmethod
    def read(cls,
             handle: Union['StreamHandle'],
             offset: Optional['int'] = None,
             size: Optional['int'] = None,
             ):
        """Read a chunk of the stream
        :param handle: Handle of the stream to read.
        :type handle: StreamHandle
        :param offset: Seek to the specified offset before reading (if not specificed, proceed with offset
following the last read). Some types of streams may only support sequential reads.
        :type offset: int
        :param size: Maximum number of bytes to read (left upon the agent discretion if not specified).
        :type size: int
        """
        return (
            cls.build_send_payload("read", {
                "handle": handle,
                "offset": offset,
                "size": size,
            }),
            cls.convert_payload({
                "base64Encoded": {
                    "class": bool,
                    "optional": True
                },
                "data": {
                    "class": str,
                    "optional": False
                },
                "eof": {
                    "class": bool,
                    "optional": False
                },
            })
        )

    @classmethod
    def resolveBlob(cls,
                    objectId: Union['Runtime.RemoteObjectId'],
                    ):
        """Return UUID of Blob object specified by a remote object id.
        :param objectId: Object id of a Blob object wrapper.
        :type objectId: Runtime.RemoteObjectId
        """
        return (
            cls.build_send_payload("resolveBlob", {
                "objectId": objectId,
            }),
            cls.convert_payload({
                "uuid": {
                    "class": str,
                    "optional": False
                },
            })
        )




# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)
from chromewhip.protocol import page as Page
from chromewhip.protocol import runtime as Runtime

# ScreenshotParams: Encoding options for a screenshot.
class ScreenshotParams(ChromeTypeBase):
    def __init__(self,
                 format: Optional['str'] = None,
                 quality: Optional['int'] = None,
                 ):

        self.format = format
        self.quality = quality


class HeadlessExperimental(PayloadMixin):
    """ This domain provides experimental commands only supported in headless mode.
    """
    @classmethod
    def beginFrame(cls,
                   frameTimeTicks: Optional['float'] = None,
                   interval: Optional['float'] = None,
                   noDisplayUpdates: Optional['bool'] = None,
                   screenshot: Optional['ScreenshotParams'] = None,
                   ):
        """Sends a BeginFrame to the target and returns when the frame was completed. Optionally captures a
screenshot from the resulting frame. Requires that the target was created with enabled
BeginFrameControl. Designed for use with --run-all-compositor-stages-before-draw, see also
https://goo.gl/3zHXhB for more background.
        :param frameTimeTicks: Timestamp of this BeginFrame in Renderer TimeTicks (milliseconds of uptime). If not set,
the current time will be used.
        :type frameTimeTicks: float
        :param interval: The interval between BeginFrames that is reported to the compositor, in milliseconds.
Defaults to a 60 frames/second interval, i.e. about 16.666 milliseconds.
        :type interval: float
        :param noDisplayUpdates: Whether updates should not be committed and drawn onto the display. False by default. If
true, only side effects of the BeginFrame will be run, such as layout and animations, but
any visual updates may not be visible on the display or in screenshots.
        :type noDisplayUpdates: bool
        :param screenshot: If set, a screenshot of the frame will be captured and returned in the response. Otherwise,
no screenshot will be captured. Note that capturing a screenshot can fail, for example,
during renderer initialization. In such a case, no screenshot data will be returned.
        :type screenshot: ScreenshotParams
        """
        return (
            cls.build_send_payload("beginFrame", {
                "frameTimeTicks": frameTimeTicks,
                "interval": interval,
                "noDisplayUpdates": noDisplayUpdates,
                "screenshot": screenshot,
            }),
            cls.convert_payload({
                "hasDamage": {
                    "class": bool,
                    "optional": False
                },
                "screenshotData": {
                    "class": str,
                    "optional": True
                },
            })
        )

    @classmethod
    def disable(cls):
        """Disables headless events for the target.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables headless events for the target.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )



class NeedsBeginFramesChangedEvent(BaseEvent):

    js_name = 'Headlessexperimental.needsBeginFramesChanged'
    hashable = []
    is_hashable = False

    def __init__(self,
                 needsBeginFrames: Union['bool', dict],
                 ):
        if isinstance(needsBeginFrames, dict):
            needsBeginFrames = bool(**needsBeginFrames)
        self.needsBeginFrames = needsBeginFrames

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

class Inspector(PayloadMixin):
    """ 
    """
    @classmethod
    def disable(cls):
        """Disables inspector domain notifications.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enables inspector domain notifications.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )



class DetachedEvent(BaseEvent):

    js_name = 'Inspector.detached'
    hashable = []
    is_hashable = False

    def __init__(self,
                 reason: Union['str', dict],
                 ):
        if isinstance(reason, dict):
            reason = str(**reason)
        self.reason = reason

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class TargetCrashedEvent(BaseEvent):

    js_name = 'Inspector.targetCrashed'
    hashable = []
    is_hashable = False

    def __init__(self):
        pass

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')


class TargetReloadedAfterCrashEvent(BaseEvent):

    js_name = 'Inspector.targetReloadedAfterCrash'
    hashable = []
    is_hashable = False

    def __init__(self):
        pass

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')



# Always prefer setuptools over distutils
from setuptools import setup, find_packages

from os import path

here = path.abspath(path.dirname(__file__))


def readme():
    try:
        with open('README.rst') as f:
            return f.read()
    except FileNotFoundError:
        return ""

setup(
    name='chromewhip',

    version='0.2.10',

    description='asyncio driver + HTTP server for Chrome devtools protocol',
    long_description=readme(),
    # The project's main homepage.
    url='https://github.com/chuckus/chromewhip',
    download_url='https://github.com/chuckus/chromewhip/archive/v0.2.10.tar.gz',

    # Author details
    author='Charlie Smith',
    author_email='charlie@chuckus.nz',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.6',
    ],

    # What does your project relate to?
    keywords='scraping chrome scraper browser automation',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'aiohttp==2.2.0', 'websockets==3.3', 'beautifulsoup4==4.6.0', 'lxml==3.8.0',
        'pyyaml==3.12', 'Pillow==4.2.1'
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['Jinja2==2.9.6', 'jsonpatch==1.16'],
        'test': ['pytest-asyncio==0.6.9'],
    },

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'chromewhip=chromewhip:main',
        ],
    },
)



# noinspection PyPep8
# noinspection PyArgumentList

"""
AUTO-GENERATED BY `scripts/generate_protocol.py` using `data/browser_protocol.json`
and `data/js_protocol.json` as inputs! Please do not modify this file.
"""

import logging
from typing import Any, Optional, Union

from chromewhip.helpers import PayloadMixin, BaseEvent, ChromeTypeBase

log = logging.getLogger(__name__)

# Metric: Run-time execution metric.
class Metric(ChromeTypeBase):
    def __init__(self,
                 name: Union['str'],
                 value: Union['float'],
                 ):

        self.name = name
        self.value = value


class Performance(PayloadMixin):
    """ 
    """
    @classmethod
    def disable(cls):
        """Disable collecting and reporting metrics.
        """
        return (
            cls.build_send_payload("disable", {
            }),
            None
        )

    @classmethod
    def enable(cls):
        """Enable collecting and reporting metrics.
        """
        return (
            cls.build_send_payload("enable", {
            }),
            None
        )

    @classmethod
    def getMetrics(cls):
        """Retrieve current values of run-time metrics.
        """
        return (
            cls.build_send_payload("getMetrics", {
            }),
            cls.convert_payload({
                "metrics": {
                    "class": [Metric],
                    "optional": False
                },
            })
        )



class MetricsEvent(BaseEvent):

    js_name = 'Performance.metrics'
    hashable = []
    is_hashable = False

    def __init__(self,
                 metrics: Union['[Metric]', dict],
                 title: Union['str', dict],
                 ):
        if isinstance(metrics, dict):
            metrics = [Metric](**metrics)
        self.metrics = metrics
        if isinstance(title, dict):
            title = str(**title)
        self.title = title

    @classmethod
    def build_hash(cls):
        raise ValueError('Unable to build hash for non-hashable type')
